# import os
# import tarfile
# from pathlib import Path
# from typing import Callable, Optional

# import click

# from coalescenceml.constants import APP_NAME, ENV_COML_CONFIG_PATH
# from coalescenceml.io.fileio import file_exists, is_remote, open


# def create_tarfile(
#     source_dir: str,
#     output_filename: str = "output.tar.gz",
#     exclude: Optional[
#         Callable[[tarfile.TarInfo], Optional[tarfile.TarInfo]]
#     ] = None,
# ) -> None:
#     """Create compressed representation of source.

#     Args:
#         source_dir: Path to source
#         output_filename: Name of outputted gz
#         exlude: function that determines whether to exclude a file.
#     """
#     if exclude_function is None:
#         # Default is to exclude venv (and X)
#         def exclude_function(
#             tarinfo: tarfile.TarInfo,
#         ) -> Optional[tarfile.TarInfo]:
#             """Exlude files from tar.

#             Args:
#                 tarinfo: Information for zipping tat

#             Returns:
#                 tarinfo requited for exclusion.
#             """
#             filename = tarinfo.name
#             if "venv/" in filename:
#                 return None
#             else:
#                 return tarinfo

#     with tarfile.open(output_filename, "w:gz") as tar:
#         tar.add(source_dir, arcname="", filter=exclude)


# def extract_tarfile(source_tar: str, output_dir: str) -> None:
#     """Extracts all files in a compressed tar file to output_dir.

#     Args:
#         source_tar: Path to a tar compressed file.
#         output_dir: Directory where to extract.
#     """
#     if is_remote(source_tar):
#         raise NotImplementedError("Unpacking a remote tar is not available at this time.")

#     with tarfile.open(source_tar, "r:gz") as tar:
#         tar.extractall(output_dir)


# def write_file_contents_as_string(file_path: str, content: str) -> None:
#     """Write contents of file.

#     Args:
#         file_path: Path to file
#         content: content to write.
#     """
#     with open(file_path, "w") as fp:
#         fp.write(content)

# def read_file_contents_as_string(file_path: str) -> str:
#     """Read contents of file.

#     Args:
#         file_path: Path to file

#     Returns:
#         content of file
#     """
#     with open(file_path, "r") as fp:
#         txt = fp.read()
#     return txt

import fnmatch
import os
from pathlib import Path
from typing import Iterable

import click
from tfx.dsl.io.filesystem import PathType

from coalescenceml.constants import (
    APP_NAME,
    ENV_COML_CONFIG_PATH,
    REMOTE_FS_PREFIX,
)
from coalescenceml.io.fileio import (
    copy,
    exists,
    isdir,
    listdir,
    makedirs,
    mkdir,
    open,
    walk,
)


def get_global_config_directory() -> str:
    """Returns the global config directory for ZenML."""
    env_var_path = os.getenv(ENV_COML_CONFIG_PATH)
    if env_var_path:
        return str(Path(env_var_path).resolve())
    return click.get_app_dir(APP_NAME)


def write_file_contents_as_string(file_path: str, content: str) -> None:
    """Writes contents of file.
    Args:
        file_path: Path to file.
        content: Contents of file.
    """
    with open(file_path, "w") as f:
        f.write(content)


def read_file_contents_as_string(file_path: str) -> str:
    """Reads contents of file.
    Args:
        file_path: Path to file.
    """
    if not exists(file_path):
        raise FileNotFoundError(f"{file_path} does not exist!")
    return open(file_path).read()  # type: ignore[no-any-return]


def find_files(dir_path: PathType, pattern: str) -> Iterable[str]:
    """Find files in a directory that match pattern.
    Args:
        dir_path: Path to directory.
        pattern: pattern like *.png.
    Yields:
         All matching filenames if found.
    """
    for root, dirs, files in walk(dir_path):
        for basename in files:
            if fnmatch.fnmatch(convert_to_str(basename), pattern):
                filename = os.path.join(
                    convert_to_str(root), convert_to_str(basename)
                )
                yield filename


def is_remote(path: str) -> bool:
    """Returns True if path exists remotely.
    Args:
        path: Any path as a string.
    Returns:
        True if remote path, else False.
    """
    return any(path.startswith(prefix) for prefix in REMOTE_FS_PREFIX)


def create_file_if_not_exists(
    file_path: str, file_contents: str = "{}"
) -> None:
    """Creates file if it does not exist.
    Args:
        file_path: Local path in filesystem.
        file_contents: Contents of file.
    """
    full_path = Path(file_path)
    if not exists(file_path):
        create_dir_recursive_if_not_exists(str(full_path.parent))
        with open(str(full_path), "w") as f:
            f.write(file_contents)


def create_dir_if_not_exists(dir_path: str) -> None:
    """Creates directory if it does not exist.
    Args:
        dir_path: Local path in filesystem.
    """
    if not isdir(dir_path):
        mkdir(dir_path)


def create_dir_recursive_if_not_exists(dir_path: str) -> None:
    """Creates directory recursively if it does not exist.
    Args:
        dir_path: Local path in filesystem.
    """
    if not isdir(dir_path):
        makedirs(dir_path)


def resolve_relative_path(path: str) -> str:
    """Takes relative path and resolves it absolutely.
    Args:
      path: Local path in filesystem.
    Returns:
        Resolved path.
    """
    if is_remote(path):
        return path
    return str(Path(path).resolve())


def copy_dir(
    source_dir: str, destination_dir: str, overwrite: bool = False
) -> None:
    """Copies dir from source to destination.
    Args:
        source_dir: Path to copy from.
        destination_dir: Path to copy to.
        overwrite: Boolean. If false, function throws an error before overwrite.
    """
    for source_file in listdir(source_dir):
        source_path = os.path.join(source_dir, convert_to_str(source_file))
        destination_path = os.path.join(
            destination_dir, convert_to_str(source_file)
        )
        if isdir(source_path):
            if source_path == destination_dir:
                # if the destination is a subdirectory of the source, we skip
                # copying it to avoid an infinite loop.
                return
            copy_dir(source_path, destination_path, overwrite)
        else:
            create_dir_recursive_if_not_exists(
                str(Path(destination_path).parent)
            )
            copy(str(source_path), str(destination_path), overwrite)


def get_grandparent(dir_path: str) -> str:
    """Get grandparent of dir.
    Args:
        dir_path: Path to directory.
    Returns:
        The input path's parent's parent.
    """
    return Path(dir_path).parent.parent.stem


def get_parent(dir_path: str) -> str:
    """Get parent of dir.
    Args:
        dir_path: Path to directory.
    Returns:
        Parent (stem) of the dir as a string.
    """
    return Path(dir_path).parent.stem


def convert_to_str(path: PathType) -> str:
    """Converts a PathType to a str using UTF-8."""
    if isinstance(path, str):
        return path
    else:
        return path.decode("utf-8")


def is_root(path: str) -> bool:
    """Returns true if path has no parent in local filesystem.
    Args:
        path: Local path in filesystem.
    Returns:
        True if root, else False.
    """
    return Path(path).parent == Path(path)



from ast import Tuple
from pathlib import Path
from s3fs import S3FileSystem

from coalescenceml.artifact_store.base_artifact_store import BaseArtifactStore, PathType
from coalescenceml.integrations.constants import AWS_ENDPOINT_STR, AWS_ENDPOINT_URL


from typing import (
    Any,
    Callable,
    Iterable,
    List,
    Optional,
)

class AWSArtifactStore(BaseArtifactStore):
    """ Artifact Store for AWS S3""" 

    # Class Configuration
    s3fs: Optional[S3FileSystem] = None
    region: Optional[str] = None

    @property 
    def filesystem(self):
        if not self.s3fs:
            if self.region:
                self.s3fs = S3FileSystem(client_kwargs={"endpoint_url": AWS_ENDPOINT_STR + self.region + AWS_ENDPOINT_URL})
            else:
                self.s3fs = S3FileSystem()
        return self.s3fs

    def open(self, name: PathType) -> Any:
        """Open a file at the given path."""
        return self.s3fs.open(name)

    def copyfile(self, src: PathType, dst: PathType, overwrite: bool = False) -> None:
        """Copy a file from the source to the destination."""
        self.s3fs.copy(src,dst)

    def exists(self, path: PathType) -> bool:
        """Returns `True` if the given path exists."""
        return self.s3fs.exists(path)

    def glob(self, pattern: PathType) -> List[PathType]:
        """Return the paths that match a glob pattern."""
        return self.s3fs.glob(pattern)

    def isdir(self, path: PathType) -> bool:
        """Returns whether the given path points to a directory."""
        return self.s3fs.isdir(path)

    def listdir(self, path: PathType) -> List[PathType]:
        """Returns a list of files under a given directory in the filesystem."""
        return self.s3fs.ls(path)

    def makedirs(self, path: PathType) -> None:
        """Make a directory at the given path, recursively creating parents."""
        self.s3fs.mkdir(path, createParents = False)

    def mkdir(self, path: PathType) -> None:
        """Make a directory at the given path; parent directory must exist."""
        self.s3fs.mkdir(path)

    def remove(self, path: PathType) -> None:
        """Remove the file at the given path. Dangerous operation."""
        self.s3fs.rm(path)

    def rename(self, src: PathType, dst: PathType, overwrite: bool = False) -> None:
        """Rename source file to destination file.
        Args:
            src: The path of the file to rename.
            dst: The path to rename the source file to.
            overwrite: If a file already exists at the destination, this
                method will overwrite it if overwrite=`True`
        """
        if not overwrite and self.exists(dst):
            raise FileExistsError(
                f"Destination path {str(dst)} already exists and argument "
                f"`overwrite` is false."
            )
        self.s3fs.rename(src, dst)

    def rmtree(self, path: PathType) -> None:
        """Deletes dir recursively. Dangerous operation."""
        self.s3fs.rm(path)

    def stat(self, path: PathType) -> Any:
        """Return the stat descriptor for a given file path."""
        return self.s3fs.info(path)

    def walk(
        self, 
        top: PathType,
        topdown: bool = True,
        onerror: Optional[Callable[..., None]] = None,
    ) -> Iterable[Tuple[PathType, List[PathType], List[PathType]]]:
        """Return an iterator that walks the contents of the given directory.
        Args:
            top: Path of directory to walk.
            topdown: Whether to walk directories topdown or bottom-up.
            onerror: Callable that gets called if an error occurs.
        Returns:
            An Iterable of Tuples, each of which contain the path of the
            current directory path, a list of directories inside the
            current directory and a list of files inside the current
            directory.
        """
        iter = self.s3fs.walk(top)	
        if topdown:
            return iter
        else:
            return reversed(list(iter))

 



class StackExistsError(Exception):
    """Raised when trying to register a stack with a name that already
    exists."""


class StackComponentExistsError(Exception):
    """Raised when trying to register a stack component with a name that
    already exists."""


class StackValidationError(Exception):
    """Raised when a stack configuration is not valid."""


class StackComponentInterfaceError(Exception):
    """Raises exception when interacting with the stack components
    in an unsupported way."""


class ProvisioningError(Exception):
    """Raised when an error occurs when provisioning resources for a
    StackComponent."""



import time

import click

from coalescenceml.cli.cli import cli
import coalescenceml.cli.utils as cli_utils
from coalescenceml.constants import console
from coalescenceml.logger import get_logger


@cli.command(name="over", help="Funny joke...hehe", hidden=True)
def print_deez_nutz_joke()->None:
    deez_nutz = r"""
      -                                                                                     -
    _/-\_                                                                                 _/-\_
 .-`-:-:-`-.   _____    ______   ______   ______    _   _   _    _   _______   ______  .-`-:-:-`-.
/-:-:-:-:-:-\ |  __ \  |  ____| |  ____| |___  /   | \ | | | |  | | |__   __| |___  / /-:-:-:-:-:-\
\:-:-:-:-:-:/ | |  | | | |__    | |__       / /    |  \| | | |  | |    | |       / /  \:-:-:-:-:-:/
 |`       `|  | |  | | |  __|   |  __|     / /     | . ` | | |  | |    | |      / /    |`       `|
 |         |  | |__| | | |____  | |____   / /__    | |\  | | |__| |    | |     / /__   |         |
 `\       /`  |_____/  |______| |______| /_____|   |_| \_|  \____/     |_|    /_____|  `\       /`
   `-._.-'                                                                               `-._.-'"""

    max_width = console.width
    deez_nutz_width = console.measure(deez_nutz).maximum
    styling = "bold cyan"
    console.print("coml over...", style="cyan", justify="left")
    with console.status("", spinner="monkey"):
        time.sleep(5)

    if deez_nutz_width >= max_width:
        console.print(deez_nutz, style=styling, overflow="crop")
        console.rule(f"")
    else:
        width_buffer = int((max_width - deez_nutz_width) / (2))
        deez_nutz = "\n".join([
            ((" "* width_buffer) + line) for line in deez_nutz.split("\n")
        ])
        console.print(deez_nutz, style=styling)
        console.rule("")



"""
CoML CLI
==================

Description of CLI

How to use the CLI
------------------

Beginning a Project
-------------------

Loading and using pre-built examples
------------------------------------
"""

from coalescenceml.cli.config import *
from coalescenceml.cli.core import *
from coalescenceml.cli.integration import *

# from coalescenceml.cli.pipeline import *
from coalescenceml.cli.stack import *
from coalescenceml.cli.stack_components import *
from coalescenceml.cli.version import *
from coalescenceml.cli.hidden import *



import datetime as dt
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

from sqlalchemy.engine.url import make_url
from sqlalchemy.exc import ArgumentError, NoResultFound
from sqlmodel import Field, Session, SQLModel, create_engine, select

from coalescenceml.enums import DirectoryStoreFlavor, StackComponentFlavor
from coalescenceml.io import utils
from coalescenceml.logger import get_logger
from coalescenceml.stack.exceptions import StackComponentExistsError
from coalescenceml.stack_store import BaseStackStore
from coalescenceml.stack_store.model import StackComponentWrapper


logger = get_logger(__name__)


class CoalescenceUser(SQLModel, table=True):
    id: int = Field(primary_key=True)
    name: str


class CoalescenceStack(SQLModel, table=True):
    name: str = Field(primary_key=True)
    created_by: int
    create_time: Optional[dt.datetime] = Field(default_factory=dt.datetime.now)


class CoalescenceStackComponent(SQLModel, table=True):
    component_type: StackComponentFlavor = Field(primary_key=True)
    name: str = Field(primary_key=True)
    component_flavor: str
    configuration: bytes  # e.g. base64 encoded json string


class CoalescenceStackDefinition(SQLModel, table=True):
    """Join table between Stacks and StackComponents"""

    stack_name: str = Field(
        primary_key=True, foreign_key="coalescencestack.name"
    )
    component_flavor: StackComponentFlavor = Field(
        primary_key=True,
        foreign_key="coalescencestackcomponent.component_flavor",
    )
    component_name: str = Field(
        primary_key=True, foreign_key="coalescencestackcomponent.name"
    )


class SqlStackStore(BaseStackStore):
    """Repository Implementation that uses SQL database backend"""

    def initialize(
        self,
        url: str,
        *args: Any,
        **kwargs: Any,
    ) -> "SqlStackStore":
        """Initialize a new SqlStackStore.

        Args:
            url: odbc path to a database.
            args, kwargs: additional parameters for SQLModel.
        Returns:
            The initialized stack store instance.
        """
        if not self.is_valid_url(url):
            raise ValueError(f"Invalid URL for SQL store: {url}")

        logger.debug("Initializing SqlStackStore at %s", url)
        self._url = url

        local_path = self.get_path_from_url(url)
        if local_path:
            utils.create_dir_recursive_if_not_exists(str(local_path.parent))

        self.engine = create_engine(url, *args, **kwargs)
        SQLModel.metadata.create_all(self.engine)
        with Session(self.engine) as session:
            if not session.exec(select(CoalescenceUser)).first():
                session.add(CoalescenceUser(id=1, name="LocalCoalescenceUser"))
            session.commit()

        super().initialize(url, *args, **kwargs)
        return self

    # Public interface implementations:

    @property
    def type(self) -> DirectoryStoreFlavor:
        """The type of stack store."""
        return DirectoryStoreFlavor.SQL

    @property
    def url(self) -> str:
        """URL of the repository."""
        if not self._url:
            raise RuntimeError(
                "SQL stack store has not been initialized. Call `initialize` "
                "before using the store."
            )
        return self._url

    @staticmethod
    def get_path_from_url(url: str) -> Optional[Path]:
        """Get the local path from a URL, if it points to a local sqlite file.

        This method first checks that the URL is a valid SQLite URL, which is
        backed by a file in the local filesystem. All other types of supported
        SQLAlchemy connection URLs are considered non-local and won't return
        a valid local path.

        Args:
            url: The URL to get the path from.

        Returns:
            The path extracted from the URL, or None, if the URL does not
            point to a local sqlite file.
        """
        if not SqlStackStore.is_valid_url(url):
            raise ValueError(f"Invalid URL for SQL store: {url}")
        if not url.startswith("sqlite:///"):
            return None
        url = url.replace("sqlite:///", "")
        return Path(url)

    @staticmethod
    def get_local_url(path: str) -> str:
        """Get a local SQL url for a given local path."""
        return f"sqlite:///{path}/coalescenceml.db"

    @staticmethod
    def is_valid_url(url: str) -> bool:
        """Check if the given url is a valid SQL url."""
        try:
            make_url(url)
        except ArgumentError:
            logger.debug("Invalid SQL URL: %s", url)
            return False

        return True

    @property
    def is_empty(self) -> bool:
        """Check if the stack store is empty."""
        with Session(self.engine) as session:
            return session.exec(select(CoalescenceStack)).first() is None

    def get_stack_configuration(
        self, name: str
    ) -> Dict[StackComponentFlavor, str]:
        """Fetches a stack configuration by name.

        Args:
            name: The name of the stack to fetch.

        Returns:
            Dict[StackComponentFlavor, str] for the requested stack name.

        Raises:
            KeyError: If no stack exists for the given name.
        """
        logger.debug("Fetching stack with name '%s'.", name)
        # first check that the stack exists
        with Session(self.engine) as session:
            maybe_stack = session.exec(
                select(CoalescenceStack).where(CoalescenceStack.name == name)
            ).first()
        if maybe_stack is None:
            raise KeyError(
                f"Unable to find stack with name '{name}'. Available names: "
                f"{set(self.stack_names)}."
            )
        # then get all components assigned to that stack
        with Session(self.engine) as session:
            definitions_and_components = session.exec(
                select(CoalescenceStackDefinition, CoalescenceStackComponent)
                .where(
                    CoalescenceStackDefinition.component_flavor
                    == CoalescenceStackComponent.component_type
                )
                .where(
                    CoalescenceStackDefinition.component_name
                    == CoalescenceStackComponent.name
                )
                .where(CoalescenceStackDefinition.stack_name == name)
            )
            params = {
                component.component_flavor: component.name
                for _, component in definitions_and_components
            }
        return {StackComponentFlavor(typ): name for typ, name in params.items()}

    @property
    def stack_configurations(
        self,
    ) -> Dict[str, Dict[StackComponentFlavor, str]]:
        """Configuration for all stacks registered in this stack store.

        Returns:
            Dictionary mapping stack names to Dict[StackComponentFlavor, str]
        """
        return {n: self.get_stack_configuration(n) for n in self.stack_names}

    def register_stack_component(
        self,
        component: StackComponentWrapper,
    ) -> None:
        """Register a stack component.

        Args:
            component: The component to register.

        Raises:
            StackComponentExistsError: If a stack component with the same type
                and name already exists.
        """
        with Session(self.engine) as session:
            existing_component = session.exec(
                select(CoalescenceStackComponent)
                .where(CoalescenceStackComponent.name == component.name)
                .where(
                    CoalescenceStackComponent.component_type == component.type
                )
            ).first()
            if existing_component is not None:
                raise StackComponentExistsError(
                    f"Unable to register stack component (type: "
                    f"{component.type}) with name '{component.name}': Found "
                    f"existing stack component with this name."
                )
            new_component = CoalescenceStackComponent(
                component_type=component.type,
                name=component.name,
                component_flavor=component.flavor,
                configuration=component.config,
            )
            session.add(new_component)
            session.commit()

    def deregister_stack(self, name: str) -> None:
        """Delete a stack from storage.

        Args:
            name: The name of the stack to be deleted.

        Raises:
            KeyError: If no stack exists for the given name.
        """
        with Session(self.engine) as session:
            try:
                stack = session.exec(
                    select(CoalescenceStack).where(
                        CoalescenceStack.name == name
                    )
                ).one()
                session.delete(stack)
            except NoResultFound as error:
                raise KeyError from error
            definitions = session.exec(
                select(CoalescenceStackDefinition).where(
                    CoalescenceStackDefinition.stack_name == name
                )
            ).all()
            for definition in definitions:
                session.delete(definition)
            session.commit()

    # Private interface implementations:

    def _create_stack(
        self, name: str, stack_configuration: Dict[StackComponentFlavor, str]
    ) -> None:
        """Add a stack to storage.

        Args:
            name: The name to save the stack as.
            stack_configuration: Dict[StackComponentFlavor, str] to persist.
        """
        with Session(self.engine) as session:
            stack = CoalescenceStack(name=name, created_by=1)
            session.add(stack)
            for ctype, cname in stack_configuration.items():
                if cname is not None:
                    session.add(
                        CoalescenceStackDefinition(
                            stack_name=name,
                            component_flavor=ctype,
                            component_name=cname,
                        )
                    )
            session.commit()

    def _get_component_flavor_and_config(
        self, component_flavor: StackComponentFlavor, name: str
    ) -> Tuple[str, bytes]:
        """Fetch the flavor and configuration for a stack component.

        Args:
            component_flavor: The type of the component to fetch.
            name: The name of the component to fetch.

        Returns:
            Pair of (flavor, configuration) for stack component, as string and
            base64-encoded yaml document, respectively

        Raises:
            KeyError: If no stack component exists for the given type and name.
        """
        with Session(self.engine) as session:
            component = session.exec(
                select(CoalescenceStackComponent)
                .where(
                    CoalescenceStackComponent.component_type == component_flavor
                )
                .where(CoalescenceStackComponent.name == name)
            ).one_or_none()
            if component is None:
                raise KeyError(
                    f"Unable to find stack component (type: {component_flavor}) "
                    f"with name '{name}'."
                )
        return component.component_flavor, component.configuration

    def _get_stack_component_names(
        self, component_flavor: StackComponentFlavor
    ) -> List[str]:
        """Get names of all registered stack components of a given type.

        Args:
            component_flavor: The type of the component to list names for.

        Returns:
            A list of names as strings.
        """
        with Session(self.engine) as session:
            statement = select(CoalescenceStackComponent).where(
                CoalescenceStackComponent.component_type == component_flavor
            )
            return [component.name for component in session.exec(statement)]

    def _delete_stack_component(
        self, component_flavor: StackComponentFlavor, name: str
    ) -> None:
        """Remove a StackComponent from storage.

        Args:
            component_flavor: The type of component to delete.
            name: Then name of the component to delete.

        Raises:
            KeyError: If no component exists for given type and name.
        """
        with Session(self.engine) as session:
            component = session.exec(
                select(CoalescenceStackComponent)
                .where(
                    CoalescenceStackComponent.component_type == component_flavor
                )
                .where(CoalescenceStackComponent.name == name)
            ).first()
            if component is not None:
                session.delete(component)
                session.commit()
            else:
                raise KeyError(
                    "Unable to deregister stack component (type: "
                    f"{component_flavor.value}) with name '{name}': No stack "
                    "component exists with this name."
                )

    # Implementation-specific internal methods:

    @property
    def stack_names(self) -> List[str]:
        """Names of all stacks registered in this StackStore."""
        with Session(self.engine) as session:
            return [s.name for s in session.exec(select(CoalescenceStack))]



from typing import Optional

import click
from rich.markdown import Markdown

from coalescenceml.cli import utils as cli_utils
from coalescenceml.cli.cli import cli
from coalescenceml.config.base_config import BaseConfiguration
from coalescenceml.config.global_config import GlobalConfiguration
from coalescenceml.config.profile_config import (
    ProfileConfiguration,
    get_default_store_type,
)
from coalescenceml.constants import console
from coalescenceml.directory import Directory
from coalescenceml.enums import DirectoryStoreFlavor, LoggingLevels


# Logging
@cli.group()
def logging() -> None:
    """Configuration of logging for ZenML pipelines."""


# Setting logging
@logging.command("set-verbosity")
@click.argument(
    "verbosity",
    type=click.Choice(
        list(map(lambda x: x.name, LoggingLevels)), case_sensitive=False
    ),
)
def set_logging_verbosity(verbosity: str) -> None:
    """Set logging level"""
    verbosity = verbosity.upper()
    if verbosity not in LoggingLevels.__members__:
        raise KeyError(
            f"Verbosity must be one of {list(LoggingLevels.__members__.keys())}"
        )
    cli_utils.info(f"Set verbosity to: {verbosity}")


# Profiles
@cli.group()
def profile() -> None:
    """Configuration of ZenML profiles."""


@profile.command("create")
@click.argument(
    "name",
    type=str,
    required=True,
)
@click.option(
    "--url",
    "-u",
    "url",
    help="The store URL to use for the profile.",
    required=False,
    type=str,
)
@click.option(
    "--store-type",
    "-t",
    "store_type",
    help="The store type to use for the profile.",
    required=False,
    type=click.Choice(list(DirectoryStoreFlavor)),
    default=get_default_store_type(),
)
def create_profile_command(
    name: str, url: Optional[str], store_type: Optional[DirectoryStoreFlavor]
) -> None:
    """Create a new configuration profile."""

    cli_utils.print_active_profile()

    cfg = GlobalConfiguration()

    if cfg.get_profile(name):
        cli_utils.error(f"Profile {name} already exists.")
        return
    cfg.add_or_update_profile(
        ProfileConfiguration(name=name, store_url=url, store_type=store_type)
    )
    cli_utils.info(f"Profile '{name}' successfully created.")


@profile.command("list")
def list_profiles_command() -> None:
    """List configuration profiles."""

    cli_utils.print_active_profile()

    cfg = GlobalConfiguration()
    dir_ = Directory()

    profiles = cfg.profiles

    if len(profiles) == 0:
        cli_utils.warning("No profiles configured!")
        return

    profile_dicts = []
    for profile_name, profile in profiles.items():
        is_active = profile_name == dir_.active_profile_name
        profile_config = {
            "ACTIVE": ":point_right:" if is_active else "",
            "PROFILE NAME": profile_name,
            "STORE TYPE": profile.store_type.value,
            "URL": profile.store_url,
            "ACTIVE STACK": dir_.active_stack_name,
        }
        profile_dicts.append(profile_config)

    cli_utils.print_table(profile_dicts)


@profile.command(
    "describe",
    help="Show details about the active profile.",
)
@click.argument(
    "name",
    type=click.STRING,
    required=False,
)
def describe_profile(name: Optional[str]) -> None:
    """Show details about a named profile or the active profile."""
    cli_utils.print_active_profile()

    dir_ = Directory()
    name = name or dir_.active_profile_name

    profile = GlobalConfiguration().get_profile(name)
    if not profile:
        cli_utils.error(f"Profile '{name}' does not exist.")
        return

    cli_utils.print_profile(
        profile,
        active=name == dir_.active_profile_name,
    )


@profile.command("delete")
@click.argument("name", type=str)
def delete_profile(name: str) -> None:
    """Delete a profile.
    If the profile is currently active, it cannot be deleted.
    """
    cli_utils.print_active_profile()

    with console.status(f"Deleting profile '{name}'...\n"):

        cfg = GlobalConfiguration()
        dir_ = Directory()
        if not cfg.get_profile(name):
            cli_utils.error(f"Profile {name} doesn't exist.")
            return
        if cfg.active_profile_name == name:
            cli_utils.error(
                f"Profile '{name}' cannot be deleted because it's globally "
                f" active. Please choose a different active global profile "
                f"first by running 'coml profile set --global PROFILE'."
            )
            return

        if dir_.active_profile_name == name:
            cli_utils.error(
                f"Profile '{name}' cannot be deleted because it's locally "
                f"active. Please choose a different active profile first by "
                f"running 'coml profile set PROFILE'."
            )
            return

        cfg.delete_profile(name)
        cli_utils.info(f"Deleted profile '{name}'.")


@profile.command("set")
@click.argument("name", type=str)
@click.option(
    "--global",
    "-g",
    "global_profile",
    is_flag=True,
    help="Set the global active profile",
)
def set_active_profile(name: str, global_profile: bool = False) -> None:
    """Set a profile as active.
    If the '--global' flag is set, the profile will be set as the global
    active profile, otherwise as the directory local active profile.
    """
    cli_utils.print_active_profile()
    scope = " global" if global_profile else ""

    with console.status(f"Setting the{scope} active profile to '{name}'..."):

        cfg: BaseConfiguration = (
            GlobalConfiguration() if global_profile else Directory()
        )

        current_profile_name = cfg.active_profile_name
        if current_profile_name == name:
            cli_utils.info(f"Profile '{name}' is already{scope}ly active.")
            return

        try:
            cfg.activate_profile(name)
        except Exception as e:
            cli_utils.error(f"Error activating{scope} profile: {str(e)}. ")
            if current_profile_name:
                cli_utils.info(
                    f"Keeping current{scope} profile: '{current_profile_name}'."
                )
                cfg.activate_profile(current_profile_name)
            return
        cli_utils.info(f"Active{scope} profile changed to: '{name}'")


@profile.command("get")
def get_active_profile() -> None:
    """Get the active profile."""
    with console.status("Getting the active profile..."):
        cli_utils.info(f"Active profile is: {Directory().active_profile_name}")


@profile.command("explain")
def explain_profile() -> None:
    """Explains the concept of CoalescenceML profiles."""

    with console.pager():
        console.print(
            Markdown(
                """
TODO: Explain in words and with examples about what a profile is and the CLI
command. This should showcase one of every command probably and the most common
methods that they will want to use.
"""
            )
        )



import click

from coalescenceml import __version__
from coalescenceml.logger import set_root_verbosity


@click.group()
@click.version_option(__version__, "--version", "-v")
def cli() -> None:
    """Coalescence ML CLI."""
    set_root_verbosity()


if __name__ == "__main__":
    cli()



import typing

import numpy as np
import sklearn as sk
from sklearn.svm import SVC
from sklearn.neighbors import KNeighborsClassifier
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn import tree
from sklearn.ensemble import RandomForestClassifier
from sklearn.naive_bayes import GaussianNB
from sklearn.cluster import KMeans
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from sklearn.model_selection import train_test_split

from coalescenceml.step import BaseStep
from coalescenceml.step import BaseStepConfig


class SKLearnTrainConfig(BaseStepConfig):
    model_name: str
    hyperparams: typing.Dict[str, typing.Any] = {}


class SKLearnTrainStep(BaseStep):
    """DOCSTRING"""

    def entrypoint(
        self,
        config: SKLearnTrainConfig,
        x: np.ndarray,
        y: np.ndarray,
    ) -> BaseEstimator:
        # Possibly change model_type to a classifier model instead of string
        # Possibly log experimental results
        # Possibly add validation data as optional too

        # Assert x and y have the same number of inputs
        assert x.shape[0] == y.shape[0]

        model_name = config.model_name
        hyperparams = config.hyperparams

        # Create model based on model_name input
        if model_name == "knn":
            model = KNeighborsClassifier(**hyperparams)
        elif model_name == "svm":
            model = SVC(**hyperparams)
        elif model_name == "linear_reg":
            model = LinearRegression(**hyperparams)
        elif model_name == "logistic_reg":
            model = LogisticRegression(**hyperparams)
        elif model_name == "decision_tree":
            model = tree.DecisionTreeClassifier(**hyperparams)
        elif model_name == "random_forest":
            model = RandomForestClassifier(**hyperparams)
        elif model_name == "gaussian_nb":
            model = GaussianNB(**hyperparams)

        # Train model
        model.fit(x, y)

        return model



import base64
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import yaml

from coalescenceml.enums import DirectoryStoreFlavor, StackComponentFlavor
from coalescenceml.logger import get_logger
from coalescenceml.stack import Stack
from coalescenceml.stack.exceptions import (
    StackComponentExistsError,
    StackExistsError,
)
from coalescenceml.stack_store.model import StackComponentWrapper, StackWrapper


logger = get_logger(__name__)


class BaseStackStore(ABC):
    """Base class for accessing data in CoalescenceML Repository and new Service."""

    def initialize(
        self,
        url: str,
        skip_default_stack: bool = False,
        *args: Any,
        **kwargs: Any,
    ) -> "BaseStackStore":
        """Initialize the store.
        Args:
            url: The URL of the store.
            skip_default_stack: If True, the creation of the default stack will
                be skipped.
            *args: Additional arguments to pass to the concrete store
                implementation.
            **kwargs: Additional keyword arguments to pass to the concrete
                store implementation.
        Returns:
            The initialized concrete store instance.
        """

        if not skip_default_stack and self.is_empty:

            logger.info("Initializing store...")
            self.register_default_stack()

        return self

    # Statics:

    @staticmethod
    @abstractmethod
    def get_path_from_url(url: str) -> Optional[Path]:
        """Get the path from a URL, if it points or is backed by a local file.
        Args:
            url: The URL to get the path from.
        Returns:
            The local path backed by the URL, or None if the URL is not backed
            by a local file or directory
        """

    @staticmethod
    @abstractmethod
    def get_local_url(path: str) -> str:
        """Get a local URL for a given local path.
        Args:
             path: the path string to build a URL out of.
        Returns:
            Url pointing to the path for the store type.
        """

    @staticmethod
    @abstractmethod
    def is_valid_url(url: str) -> bool:
        """Check if the given url is valid."""

    # Public Interface:

    @property
    @abstractmethod
    def type(self) -> DirectoryStoreFlavor:
        """The type of stack store."""

    @property
    @abstractmethod
    def url(self) -> str:
        """Get the repository URL."""

    @property
    @abstractmethod
    def is_empty(self) -> bool:
        """Check if the store is empty (no stacks are configured).
        The implementation of this method should check if the store is empty
        without having to load all the stacks from the persistent storage.
        """

    @abstractmethod
    def get_stack_configuration(
        self, name: str
    ) -> Dict[StackComponentFlavor, str]:
        """Fetches a stack configuration by name.
        Args:
            name: The name of the stack to fetch.
        Returns:
            Dict[StackComponentFlavor, str] for the requested stack name.
        Raises:
            KeyError: If no stack exists for the given name.
        """

    @property
    @abstractmethod
    def stack_configurations(
        self,
    ) -> Dict[str, Dict[StackComponentFlavor, str]]:
        """Configurations for all stacks registered in this stack store.
        Returns:
            Dictionary mapping stack names to Dict[StackComponentFlavor, str]'s
        """

    @abstractmethod
    def register_stack_component(
        self,
        component: StackComponentWrapper,
    ) -> None:
        """Register a stack component.
        Args:
            component: The component to register.
        Raises:
            StackComponentExistsError: If a stack component with the same type
                and name already exists.
        """

    @abstractmethod
    def deregister_stack(self, name: str) -> None:
        """Delete a stack from storage.
        Args:
            name: The name of the stack to be deleted.
        Raises:
            KeyError: If no stack exists for the given name.
        """

    # Private interface (must be implemented, not to be called by user):

    @abstractmethod
    def _create_stack(
        self, name: str, stack_configuration: Dict[StackComponentFlavor, str]
    ) -> None:
        """Add a stack to storage.
        Args:
            name: The name to save the stack as.
            stack_configuration: Dict[StackComponentFlavor, str] to persist.
        """

    @abstractmethod
    def _get_component_flavor_and_config(
        self, component_flavor: StackComponentFlavor, name: str
    ) -> Tuple[str, bytes]:
        """Fetch the flavor and configuration for a stack component.
        Args:
            component_flavor: The type of the component to fetch.
            name: The name of the component to fetch.
        Returns:
            Pair of (flavor, configuration) for stack component, as string and
            base64-encoded yaml document, respectively
        Raises:
            KeyError: If no stack component exists for the given type and name.
        """

    @abstractmethod
    def _get_stack_component_names(
        self, component_flavor: StackComponentFlavor
    ) -> List[str]:
        """Get names of all registered stack components of a given type.
        Args:
            component_flavor: The type of the component to list names for.
        Returns:
            A list of names as strings.
        """

    @abstractmethod
    def _delete_stack_component(
        self, component_flavor: StackComponentFlavor, name: str
    ) -> None:
        """Remove a StackComponent from storage.
        Args:
            component_flavor: The type of component to delete.
            name: Then name of the component to delete.
        Raises:
            KeyError: If no component exists for given type and name.
        """

    # Common code (user facing):

    @property
    def stacks(self) -> List[StackWrapper]:
        """All stacks registered in this stack store."""
        return [
            self._stack_from_dict(name, conf)
            for name, conf in self.stack_configurations.items()
        ]

    def get_stack(self, name: str) -> StackWrapper:
        """Fetch a stack by name.
        Args:
            name: The name of the stack to retrieve.
        Returns:
            StackWrapper instance if the stack exists.
        Raises:
            KeyError: If no stack exists for the given name.
        """
        return self._stack_from_dict(name, self.get_stack_configuration(name))

    def register_stack(self, stack: StackWrapper) -> Dict[str, str]:
        """Register a stack and its components.
        If any of the stacks' components aren't registered in the stack store
        yet, this method will try to register them as well.
        Args:
            stack: The stack to register.
        Returns:
            metadata dict for telemetry or logging.
        Raises:
            StackExistsError: If a stack with the same name already exists.
            StackComponentExistsError: If a component of the stack wasn't
                registered and a different component with the same name
                already exists.
        """
        try:
            self.get_stack(stack.name)
        except KeyError:
            pass
        else:
            raise StackExistsError(
                f"Unable to register stack with name '{stack.name}': Found "
                f"existing stack with this name."
            )

        def __check_component(
            component: StackComponentWrapper,
        ) -> Tuple[StackComponentFlavor, str]:
            """Try to register a stack component, if it doesn't exist.
            Args:
                component: StackComponentWrapper to register.
            Returns:
                metadata key value pair for telemetry.
            Raises:
                StackComponentExistsError: If a component with same name exists.
            """
            try:
                existing_component = self.get_stack_component(
                    component_flavor=component.type, name=component.name
                )
                if existing_component.uuid != component.uuid:
                    raise StackComponentExistsError(
                        f"Unable to register one of the stacks components: "
                        f"A component of type '{component.type}' and name "
                        f"'{component.name}' already exists."
                    )
            except KeyError:
                self.register_stack_component(component)
            return component.type, component.name

        stack_configuration = {
            typ: name for typ, name in map(__check_component, stack.components)
        }
        metadata = {c.type.value: c.flavor for c in stack.components}
        self._create_stack(stack.name, stack_configuration)
        return metadata

    def get_stack_component(
        self, component_flavor: StackComponentFlavor, name: str
    ) -> StackComponentWrapper:
        """Get a registered stack component.
        Raises:
            KeyError: If no component with the requested type and name exists.
        """
        flavor, config = self._get_component_flavor_and_config(
            component_flavor, name=name
        )
        uuid = yaml.safe_load(base64.b64decode(config).decode())["uuid"]
        return StackComponentWrapper(
            type=component_flavor,
            flavor=flavor,
            name=name,
            uuid=uuid,
            config=config,
        )

    def get_stack_components(
        self, component_flavor: StackComponentFlavor
    ) -> List[StackComponentWrapper]:
        """Fetches all registered stack components of the given type.
        Args:
            component_flavor: StackComponentFlavor to list members of
        Returns:
            A list of StackComponentConfiguration instances.
        """
        return [
            self.get_stack_component(
                component_flavor=component_flavor, name=name
            )
            for name in self._get_stack_component_names(component_flavor)
        ]

    def deregister_stack_component(
        self, component_flavor: StackComponentFlavor, name: str
    ) -> None:
        """Deregisters a stack component.
        Args:
            component_flavor: The type of the component to deregister.
            name: The name of the component to deregister.
        Raises:
            ValueError: if trying to deregister a component that's part
                of a stack.
        """
        for stack_name, stack_config in self.stack_configurations.items():
            if stack_config.get(component_flavor) == name:
                raise ValueError(
                    f"Unable to deregister stack component (type: "
                    f"{component_flavor}, name: {name}) that is part of a "
                    f"registered stack (stack name: '{stack_name}')."
                )
        self._delete_stack_component(component_flavor, name=name)

    def register_default_stack(self) -> None:
        """Populates the store with the default Stack.
        The default stack contains a local orchestrator,
        a local artifact store and a local SQLite metadata store.
        """
        stack = Stack.default_local_stack()
        metadata = self.register_stack(StackWrapper.from_stack(stack))
        metadata["store_type"] = self.type.value

    # Common code (internal implementations, private):

    def _stack_from_dict(
        self, name: str, stack_configuration: Dict[StackComponentFlavor, str]
    ) -> StackWrapper:
        """Build a StackWrapper from stored configurations"""
        stack_components = [
            self.get_stack_component(
                component_flavor=component_flavor, name=component_name
            )
            for component_flavor, component_name in stack_configuration.items()
        ]
        return StackWrapper(name=name, components=stack_components)



from coalescenceml.integrations.constants import XGBOOST
from coalescenceml.integrations.integration import Integration


class XgboostIntegration(Integration):
    """XGBoost integration for Coalescence."""

    NAME = XGBOOST
    REQUIREMENTS = ["xgboost==1.6.0"]

    @classmethod
    def activate(cls) -> None:
        """Activate the integration."""
        from coalescenceml.integrations.xgboost import producers



import base64
import json
import os
import re
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

from coalescenceml.enums import DirectoryStoreFlavor, StackComponentFlavor
from coalescenceml.io import fileio, utils
from coalescenceml.logger import get_logger
from coalescenceml.stack.exceptions import StackComponentExistsError
from coalescenceml.stack_store import BaseStackStore
from coalescenceml.stack_store.model import (
    StackComponentWrapper,
    StackStoreModel,
)
from coalescenceml.utils import yaml_utils


logger = get_logger(__name__)


class LocalStackStore(BaseStackStore):
    def initialize(
        self,
        url: str,
        *args: Any,
        stack_data: Optional[StackStoreModel] = None,
        **kwargs: Any,
    ) -> "LocalStackStore":
        """Initializes a local stack store instance.
        Args:
            url: URL of local directory of the repository to use for
                stack storage.
            stack_data: optional stack data store object to pre-populate the
                stack store with.
            args: additional positional arguments (ignored).
            kwargs: additional keyword arguments (ignored).
        Returns:
            The initialized stack store instance.
        """
        if not self.is_valid_url(url):
            raise ValueError(f"Invalid URL for local store: {url}")

        self._root = self.get_path_from_url(url)
        self._url = f"file://{self._root}"
        utils.create_dir_recursive_if_not_exists(str(self._root))

        if stack_data is not None:
            self.__store = stack_data
            self._write_store()
        elif fileio.exists(self._store_path()):
            config_dict = yaml_utils.read_yaml(self._store_path())
            self.__store = StackStoreModel.parse_obj(config_dict)
        else:
            self.__store = StackStoreModel.empty_store()
            self._write_store()

        super().initialize(url, *args, **kwargs)
        return self

    # Public interface implementations:

    @property
    def type(self) -> DirectoryStoreFlavor:
        """The type of stack store."""
        return DirectoryStoreFlavor.LOCAL

    @property
    def url(self) -> str:
        """URL of the repository."""
        return self._url

    @staticmethod
    def get_path_from_url(url: str) -> Optional[Path]:
        """Get the path from a URL.
        Args:
            url: The URL to get the path from.
        Returns:
            The path from the URL.
        """
        if not LocalStackStore.is_valid_url(url):
            raise ValueError(f"Invalid URL for local store: {url}")
        url = url.replace("file://", "")
        return Path(url)

    @staticmethod
    def get_local_url(path: str) -> str:
        """Get a local URL for a given local path."""
        return f"file://{path}"

    @staticmethod
    def is_valid_url(url: str) -> bool:
        """Check if the given url is a valid local path."""
        scheme = re.search("^([a-z0-9]+://)", url)
        return not scheme or scheme.group() == "file://"

    @property
    def is_empty(self) -> bool:
        """Check if the stack store is empty."""
        return len(self.__store.stacks) == 0

    def get_stack_configuration(
        self, name: str
    ) -> Dict[StackComponentFlavor, str]:
        """Fetches a stack configuration by name.
        Args:
            name: The name of the stack to fetch.
        Returns:
            Dict[StackComponentFlavor, str] for the requested stack name.
        Raises:
            KeyError: If no stack exists for the given name.
        """
        logger.debug("Fetching stack with name '%s'.", name)
        if name not in self.__store.stacks:
            raise KeyError(
                f"Unable to find stack with name '{name}'. Available names: "
                f"{set(self.__store.stacks)}."
            )

        return self.__store.stacks[name]

    @property
    def stack_configurations(
        self,
    ) -> Dict[str, Dict[StackComponentFlavor, str]]:
        """Configuration for all stacks registered in this stack store.
        Returns:
            Dictionary mapping stack names to Dict[StackComponentFlavor, str]
        """
        return self.__store.stacks.copy()

    def register_stack_component(
        self,
        component: StackComponentWrapper,
    ) -> None:
        """Register a stack component.
        Args:
            component: The component to register.
        Raises:
            StackComponentExistsError: If a stack component with the same type
                and name already exists.
        """
        components = self.__store.stack_components[component.type]
        if component.name in components:
            raise StackComponentExistsError(
                f"Unable to register stack component (type: {component.type}) "
                f"with name '{component.name}': Found existing stack component "
                f"with this name."
            )

        # write the component configuration file
        component_config_path = self._get_stack_component_config_path(
            component_flavor=component.type, name=component.name
        )
        utils.create_dir_recursive_if_not_exists(
            os.path.dirname(component_config_path)
        )
        utils.write_file_contents_as_string(
            component_config_path,
            base64.b64decode(component.config).decode(),
        )

        # add the component to the stack store dict and write it to disk
        components[component.name] = component.flavor
        self._write_store()
        logger.info(
            "Registered stack component with type '%s' and name '%s'.",
            component.type,
            component.name,
        )

    def deregister_stack(self, name: str) -> None:
        """Remove a stack from storage.
        Args:
            name: The name of the stack to be deleted.
        Raises:
            KeyError: If no stack exists for the given name.
        """
        del self.__store.stacks[name]
        self._write_store()

    # Private interface implementations:

    def _create_stack(
        self, name: str, stack_configuration: Dict[StackComponentFlavor, str]
    ) -> None:
        """Add a stack to storage.
        Args:
            name: The name to save the stack as.
            stack_configuration: Dict[StackComponentFlavor, str] to persist.
        """
        self.__store.stacks[name] = stack_configuration
        self._write_store()
        logger.info("Registered stack with name '%s'.", name)

    def _get_component_flavor_and_config(
        self, component_flavor: StackComponentFlavor, name: str
    ) -> Tuple[str, bytes]:
        """Fetch the flavor and configuration for a stack component.
        Args:
            component_flavor: The type of the component to fetch.
            name: The name of the component to fetch.
        Returns:
            Pair of (flavor, configuration) for stack component, as string and
            base64-encoded yaml document, respectively
        Raises:
            KeyError: If no stack component exists for the given type and name.
        """
        components: Dict[str, str] = self.__store.stack_components[
            component_flavor
        ]
        if name not in components:
            raise KeyError(
                f"Unable to find stack component (type: {component_flavor}) "
                f"with name '{name}'. Available names: {set(components)}."
            )

        component_config_path = self._get_stack_component_config_path(
            component_flavor=component_flavor, name=name
        )
        flavor = components[name]
        config = base64.b64encode(
            utils.read_file_contents_as_string(component_config_path).encode()
        )
        return flavor, config

    def _get_stack_component_names(
        self, component_flavor: StackComponentFlavor
    ) -> List[str]:
        """Get names of all registered stack components of a given type."""
        return list(self.__store.stack_components[component_flavor])

    def _delete_stack_component(
        self, component_flavor: StackComponentFlavor, name: str
    ) -> None:
        """Remove a StackComponent from storage.
        Args:
            component_flavor: The type of component to delete.
            name: Then name of the component to delete.
        Raises:
            KeyError: If no component exists for given type and name.
        """
        component_config_path = self._get_stack_component_config_path(
            component_flavor=component_flavor, name=name
        )

        if fileio.exists(component_config_path):
            fileio.remove(component_config_path)

        components = self.__store.stack_components[component_flavor]
        del components[name]
        self._write_store()

    # Implementation-specific internal methods:

    @property
    def root(self) -> Path:
        """The root directory of the stack store."""
        if not self._root:
            raise RuntimeError(
                "Local stack store has not been initialized. Call `initialize` "
                "before using the store."
            )
        return self._root

    def _get_stack_component_config_path(
        self, component_flavor: StackComponentFlavor, name: str
    ) -> str:
        """Path to the configuration file of a stack component."""
        path = self.root / component_flavor / f"{name}.yaml"
        return str(path)

    def _store_path(self) -> str:
        """Path to the stack store yaml file."""
        return str(self.root / "stacks.yaml")

    def _write_store(self) -> None:
        """Writes the stack store yaml file."""
        config_dict = json.loads(self.__store.json())
        yaml_utils.write_yaml(self._store_path(), config_dict)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional


if TYPE_CHECKING:
    from coalescenceml.config.profile_config import ProfileConfiguration


class BaseConfiguration(ABC):
    """Base class for global configuration management.
    This class defines the common interface related to profile and stack
    management that all global configuration classes must implement.
    Both the GlobalConfiguration and Repository classes implement this class,
    since they share similarities concerning the management of active profiles
    and stacks.
    """

    @abstractmethod
    def activate_profile(self, profile_name: str) -> None:
        """Set the active profile
        Args:
            profile_name: The name of the profile to set as active.
        Raises:
            KeyError: If the profile with the given name does not exist.
        """

    @property
    @abstractmethod
    def active_profile(self) -> Optional[ProfileConfiguration]:
        """Return the profile set as active for the repository.
        Returns:
            The active profile or None, if no active profile is set.
        """

    @property
    @abstractmethod
    def active_profile_name(self) -> Optional[str]:
        """Return the name of the profile set as active.
        Returns:
            The active profile name or None, if no active profile is set.
        """

    @abstractmethod
    def activate_stack(self, stack_name: str) -> None:
        """Set the active stack for the active profile.
        Args:
            stack_name: name of the stack to activate
        Raises:
            KeyError: If the stack with the given name does not exist.
        """

    @property
    @abstractmethod
    def active_stack_name(self) -> Optional[str]:
        """Get the active stack name from the active profile.
        Returns:
            The active stack name or None if no active stack is set or if
            no active profile is set.
        """



from __future__ import annotations

import textwrap
from abc import ABC
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Optional, Set
from uuid import UUID, uuid4

from pydantic import BaseModel, Field, root_validator

from coalescenceml.enums import StackComponentFlavor
from coalescenceml.integrations.utils import get_requirements_for_module
from coalescenceml.stack.exceptions import StackComponentInterfaceError


if TYPE_CHECKING:
    from coalescenceml.pipeline import BasePipeline
    from coalescenceml.pipeline.runtime_configuration import (
        RuntimeConfiguration,
    )
    from coalescenceml.stack import Stack, StackValidator


class StackComponent(BaseModel, ABC):
    """Abstract StackComponent class for all components of a CoalescenceML stack.

    Attributes:
        name: The name of the component.
        uuid: Unique identifier of the component.
    """

    name: str
    uuid: UUID = Field(default_factory=uuid4)

    # Class Configuration
    TYPE: ClassVar[StackComponentFlavor]
    FLAVOR: ClassVar[str]

    @property
    def log_file(self) -> Optional[str]:
        """Optional path to a log file for the stack component."""
        # TODO: Add support for multiple log files for a stack
        #  component. E.g. let each component return a generator that yields
        #  logs instead of specifying a single file path.
        return None

    @property
    def runtime_options(self) -> Dict[str, Any]:
        """Runtime options that are available to configure this component.

        The items of the dictionary should map option names (which can be used
        to configure the option in the `RuntimeConfiguration`) to default
        values for the option (or `None` if there is no default value).
        """
        return {}

    @property
    def requirements(self) -> Set[str]:
        """Set of PyPI requirements for the component."""
        return set(get_requirements_for_module(self.__module__))

    def prepare_pipeline_deployment(
        self,
        pipeline: BasePipeline,
        stack: Stack,
        runtime_configuration: RuntimeConfiguration,
    ) -> None:
        """Prepares deploying the pipeline.

        This method gets called immediately before a pipeline is deployed.
        Subclasses should override it if they require runtime configuration
        options or if they need to run code before the pipeline deployment.

        Args:
            pipeline: The pipeline that will be deployed.
            stack: The stack on which the pipeline will be deployed.
            runtime_configuration: Contains all the runtime configuration
                options specified for the pipeline run.
        """

    def prepare_pipeline_run(self) -> None:
        """Prepares running the pipeline."""

    def cleanup_pipeline_run(self) -> None:
        """Cleans up resources after the pipeline run is finished."""

    def prepare_step_run(self) -> None:
        """Prepare running a step."""

    def cleanup_step_run(self) -> None:
        """Cleanup resources after the step is run."""

    @property
    def validator(self) -> Optional["StackValidator"]:
        """The optional validator of the stack component.

        This validator will be called each time a stack with the stack
        component is initialized. Subclasses should override this property
        and return a `StackValidator` that makes sure they're not included in
        any stack that they're not compatible with.
        """
        return None

    @property
    def is_provisioned(self) -> bool:
        """If the component provisioned resources to run locally."""
        return True

    @property
    def is_running(self) -> bool:
        """If the component is running locally."""
        return True

    @property
    def is_suspended(self) -> bool:
        """If the component is suspended."""
        return not self.is_running

    def provision(self) -> None:
        """Provisions resources to run the component locally."""
        raise NotImplementedError(
            f"Provisioning local resources not implemented for {self}."
        )

    def deprovision(self) -> None:
        """Deprovisions all local resources of the component."""
        raise NotImplementedError(
            f"Deprovisioning local resource not implemented for {self}."
        )

    def resume(self) -> None:
        """Resumes the provisioned local resources of the component."""
        raise NotImplementedError(
            f"Resuming provisioned resources not implemented for {self}."
        )

    def suspend(self) -> None:
        """Suspends the provisioned local resources of the component."""
        raise NotImplementedError(
            f"Suspending provisioned resources not implemented for {self}."
        )

    def __repr__(self) -> str:
        """String representation of the stack component."""
        attribute_representation = ", ".join(
            f"{key}={value}" for key, value in self.dict().items()
        )
        return (
            f"{self.__class__.__qualname__}(type={self.TYPE}, "
            f"flavor={self.FLAVOR}, {attribute_representation})"
        )

    def __str__(self) -> str:
        """String representation of the stack component."""
        return self.__repr__()

    @root_validator
    def _ensure_stack_component_complete(cls, values: Dict[str, Any]) -> Any:
        try:
            stack_component_flavor = getattr(cls, "TYPE")
            assert stack_component_flavor in StackComponentFlavor
        except (AttributeError, AssertionError):
            raise StackComponentInterfaceError(
                textwrap.dedent(
                    """
                    When you are working with any classes which subclass from
                    `coalescenceml.stack.StackComponent` please make sure that your
                    class has a ClassVar named `TYPE` and its value is set to a
                    `StackComponentFlavor` from `from coalescenceml.enums import StackComponentFlavor`.

                    In most of the cases, this is already done for you within the
                    implementation of the base concept.

                    Example:

                    class BaseArtifactStore(StackComponent):
                        # Instance Variables
                        path: str

                        # Class Variables
                        TYPE: ClassVar[StackComponentFlavor] = StackComponentFlavor.ARTIFACT_STORE
                    """
                )
            )

        try:
            getattr(cls, "FLAVOR")
        except AttributeError:
            raise StackComponentInterfaceError(
                textwrap.dedent(
                    """
                    When you are working with any classes which subclass from
                    `coalescenceml.stack.StackComponent` please make sure that your
                    class has a defined ClassVar `FLAVOR`.

                    Example:

                    class LocalArtifactStore(BaseArtifactStore):

                        ...

                        # Define flavor as a ClassVar
                        FLAVOR: ClassVar[str] = "local"

                        ...
                    """
                )
            )

        return values

    class Config:
        """Pydantic configuration class."""

        # public attributes are immutable
        allow_mutation = False
        # all attributes with leading underscore are private and therefore
        # are mutable and not included in serialization
        underscore_attrs_are_private = True



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, Type

from coalescenceml.logger import get_logger
from coalescenceml.utils import source_utils


if TYPE_CHECKING:
    from coalescenceml.metadata_store import BaseMetadataStore
    from coalescenceml.post_execution.step import StepView
    from coalescenceml.producers.base_producer import BaseProducer

logger = get_logger(__name__)


class ArtifactView:
    """Post-execution artifact class which can be used to read
    artifact data that was created during a pipeline execution.
    """

    def __init__(
        self,
        id_: int,
        type_: str,
        uri: str,
        producer: str,
        data_type: str,
        metadata_store: BaseMetadataStore,
        parent_step_id: int,
    ):
        """Initializes a post-execution artifact object.
        In most cases `ArtifactView` objects should not be created manually but
        retrieved from a `StepView` via the `inputs` or `outputs` properties.
        Args:
            id_: The artifact id.
            type_: The type of this artifact.
            uri: Specifies where the artifact data is stored.
            producer: Information needed to restore the producer
                that was used to write this artifact.
            data_type: The type of data that was passed to the producer
                when writing that artifact. Will be used as a default type
                to read the artifact.
            metadata_store: The metadata store which should be used to fetch
                additional information related to this pipeline.
            parent_step_id: The ID of the parent step.
        """
        self._id = id_
        self._type = type_
        self._uri = uri
        self._producer = producer
        self._data_type = data_type
        self._metadata_store = metadata_store
        self._parent_step_id = parent_step_id

    @property
    def id(self) -> int:
        """Returns the artifact id."""
        return self._id

    @property
    def type(self) -> str:
        """Returns the artifact type."""
        return self._type

    @property
    def data_type(self) -> str:
        """Returns the data type of the artifact."""
        return self._data_type

    @property
    def uri(self) -> str:
        """Returns the URI where the artifact data is stored."""
        return self._uri

    @property
    def parent_step_id(self) -> int:
        """Returns the ID of the parent step. This need not be equivalent to
        the ID of the producer step."""
        return self._parent_step_id

    @property
    def producer_step(self) -> StepView:
        """Returns the original StepView that produced the artifact."""
        # TODO: Replace with artifact.id instead of passing self if
        #  required.
        return self._metadata_store.get_producer_step_from_artifact(self)

    @property
    def is_cached(self) -> bool:
        """Returns True if artifact was cached in a previous run, else False."""
        # self._metadata_store.
        return self.producer_step.id != self.parent_step_id

    def read(
        self,
        output_data_type: Optional[Type[Any]] = None,
        producer_class: Optional[Type[BaseProducer]] = None,
    ) -> Any:
        """Produces the data stored in this artifact.
        Args:
            output_data_type: The datatype to which the producer should
                read, will be passed to the producers `handle_input` method.
            producer_class: The class of the producer that should be
                used to read the artifact data. If no producer class is
                given, we use the producer that was used to write the
                artifact during execution of the pipeline.
        Returns:
              The produced data.
        """

        if not producer_class:
            try:
                producer_class = source_utils.load_source_path_class(
                    self._producer
                )
            except (ModuleNotFoundError, AttributeError) as e:
                logger.error(
                    f"CoalescenceML can not locate and import the producer module "
                    f"{self._producer} which was used to write this "
                    f"artifact. If you want to read from it, please provide "
                    f"a 'producer_class'."
                )
                raise ModuleNotFoundError(e) from e

        if not output_data_type:
            try:
                output_data_type = source_utils.load_source_path_class(
                    self._data_type
                )
            except (ModuleNotFoundError, AttributeError) as e:
                logger.error(
                    f"CoalescenceML can not locate and import the data type of this "
                    f"artifact {self._data_type}. If you want to read "
                    f"from it, please provide a 'output_data_type'."
                )
                raise ModuleNotFoundError(e) from e

        logger.debug(
            "Using '%s' to read '%s' (uri: %s).",
            producer_class.__qualname__,
            self._type,
            self._uri,
        )

        # TODO: passing in `self` to initialize the producer only
        #  works because producers only require a `.uri` property at the
        #  moment.
        producer = producer_class(self)
        return producer.handle_input(output_data_type)

    def __repr__(self) -> str:
        """Returns a string representation of this artifact."""
        return (
            f"{self.__class__.__qualname__}(id={self._id}, "
            f"type='{self._type}', uri='{self._uri}', "
            f"producer='{self._producer}')"
        )

    def __eq__(self, other: Any) -> bool:
        """Returns whether the other object is referring to the
        same artifact."""
        if isinstance(other, ArtifactView):
            return self._id == other._id and self._uri == other._uri
        return False



from typing import Any, Iterator, NamedTuple, Tuple, Type


class Output:
    """Special object to store namedtuple with immutable name."""

    def __init__(self, **kwargs: Type[Any]):
        self.outputs = NamedTuple("CoalescenceOutput", **kwargs)

    def items(self) -> Iterator[Tuple[str, Type[Any]]]:
        """Items returns an iterator over the output.

        Yields:
            Tuple[str, Type[Any]]: A pair of the output name and type
        """
        yield from self.outputs.__annotations__.items()



from coalescenceml.integrations.constants import MLFLOW
from coalescenceml.integrations.integration import Integration


class MLFlowIntegration(Integration):
    """MLFlow integration for CoalescenceML."""

    NAME = MLFLOW
    REQUIREMENTS = [
        "mlflow==1.25.1",
        "mlserver==1.0.1",
        "mlserver-mlflow==1.0.1",
    ]

    @classmethod
    def activate(cls) -> None:
        from coalescenceml.integrations.mlflow import experiment_tracker


MLFlowIntegration.check_installation()



import os
from typing import Any, Type

import tensorflow as tf

from coalescenceml.artifacts import ModelArtifact
from coalescenceml.io import fileio
from coalescenceml.logger import get_logger
from coalescenceml.producers.base_producer import BaseProducer
from coalescenceml.producers.producer_registry import register_producer_class


logger = get_logger(__name__)
DEFAULT_FILENAME = "model.keras.hdf5"


@register_producer_class
class KerasModelProducer(BaseProducer):
    """Read/Write Keras Model files."""

    ARTIFACT_TYPES = (ModelArtifact,)
    TYPES = (tf.keras.Model,)

    def handle_input(self, data_type: Type[Any]) -> tf.keras.Model:
        """Reads keras model from h5 file."""
        super().handle_input(data_type)
        filepath = os.path.join(self.artifact.uri, DEFAULT_FILENAME)
        contents = tf.keras.models.load_model(filepath)
        return contents

    def handle_return(self, model: tf.keras.Model) -> None:
        """Writes a keras model to artifact store as h5"""
        super().handle_return(model)
        filepath = os.path.join(self.artifact.uri, DEFAULT_FILENAME)
        model.save(filepath)



from typing import Any, Callable, Dict, Optional, Type, TypeVar, Union, overload

from coalescenceml.artifacts.base_artifact import BaseArtifact
from coalescenceml.step import BaseStep
from coalescenceml.step.utils import (
    INSTANCE_CONFIGURATION,
    OUTPUT_SPEC,
    PARAM_CREATED_BY_FUNCTIONAL_API,
    PARAM_CUSTOM_STEP_OPERATOR,
    PARAM_ENABLE_CACHE,
    STEP_INNER_FUNC_NAME,
)


F = TypeVar("F", bound=Callable[..., Any])


@overload
def step(func: F) -> Type[BaseStep]:
    ...

def step(
    *,
    name: Optional[str] = None,
    enable_cache: Optional[bool] = None,
    output_types: Optional[Dict[str, Type[BaseArtifact]]] = None,
    custom_step_operator: Optional[str] = None,
) -> Callable[[F], Type[BaseStep]]:
    ...

def step(
    func: F = None,
    *,
    name: Optional[str] = None,
    enable_cache: Optional[bool] = None,
    output_types: Optional[Dict[str, Type[BaseArtifact]]] = None,
    custom_step_operator: Optional[str] = None
) -> Union[Callable[[F], Type[BaseStep]], Type[BaseStep]]:
    """ """
    def step_decor(func_: F) -> Type[BaseStep]:
        step_name = name or func_.__name__
        output_spec = output_types or {}

        return type(
            step_name,
            (BaseStep,),
            {
                STEP_INNER_FUNC_NAME: staticmethod(func),
                INSTANCE_CONFIGURATION: {
                    PARAM_ENABLE_CACHE: enable_cache,
                    PARAM_CREATED_BY_FUNCTIONAL_API: True,
                    PARAM_CUSTOM_STEP_OPERATOR: custom_step_operator,
                },
                OUTPUT_SPEC: output_spec,
                "__model__": func.__module__,
            },
        )

    if func is None:
        return step_decor
    else:
        return step_decor(func)



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any, Optional

import requests
from pydantic import BaseModel, Field

from coalescenceml.constants import ENV_COML_DEFAULT_STORE_TYPE
from coalescenceml.enums import DirectoryStoreFlavor
from coalescenceml.io import fileio
from coalescenceml.logger import get_logger


if TYPE_CHECKING:
    from coalescenceml.config.global_config import GlobalConfiguration


logger = get_logger(__name__)


DEFAULT_PROFILE_NAME = "default"


def get_default_store_type() -> DirectoryStoreFlavor:
    """Return the default store type.
    The default store type can be set via the environment variable
    COML_DEFAULT_STORE_TYPE. If this variable is not set, the default
    store type is set to 'LOCAL'.
    NOTE: this is a global function instead of a default
    `ProfileConfiguration.store_type` value because it makes it easier to mock
    in the unit tests.
    Returns:
        The default store type.
    """
    store_type = os.getenv(ENV_COML_DEFAULT_STORE_TYPE)
    if store_type and store_type in DirectoryStoreFlavor.values():
        return DirectoryStoreFlavor(store_type)
    return DirectoryStoreFlavor.LOCAL


class ProfileConfiguration(BaseModel):
    """Stores configuration profile options.
    Attributes:
        name: Name of the profile.
        store_url: URL pointing to the CoalescenceML store backend.
        store_type: Type of the store backend.
        active_stack: Optional name of the active stack.
        _config: global configuration to which this profile belongs.
    """

    name: str
    store_url: Optional[str]
    store_type: DirectoryStoreFlavor = Field(
        default_factory=get_default_store_type
    )
    active_stack: Optional[str]
    _config: Optional["GlobalConfiguration"]

    def __init__(
        self, config: Optional[GlobalConfiguration] = None, **kwargs: Any
    ) -> None:
        """Initializes a ProfileConfiguration object.
        Args:
            config: global configuration to which this profile belongs. When not
                specified, the default global configuration path is used.
            **kwargs: additional keyword arguments are passed to the
                BaseModel constructor.
        """
        self._config = config
        super().__init__(**kwargs)

    @property
    def config_directory(self) -> str:
        """Directory where the profile configuration is stored."""
        return os.path.join(
            self.global_config.config_directory, "profiles", self.name
        )

    def initialize(self) -> None:
        """Initialize the profile."""

        # import here to avoid circular dependency
        from coalescenceml.directory import Directory

        logger.info("Initializing profile `%s`...", self.name)

        # Create and initialize the profile using a special directory instance.
        # This also validates and updates the store URL configuration and
        # creates all necessary resources (e.g. paths, initial DB, default
        # stacks).
        repo = Directory(profile=self)

        if not self.active_stack:
            try:
                stacks = repo.stacks
            except requests.exceptions.ConnectionError:
                stacks = None
            if stacks:
                self.active_stack = stacks[0].name

    def cleanup(self) -> None:
        """Cleanup the profile directory."""
        if fileio.isdir(self.config_directory):
            fileio.rmtree(self.config_directory)

    @property
    def global_config(self) -> GlobalConfiguration:
        """Return the global configuration to which this profile belongs."""
        from coalescenceml.config.global_config import GlobalConfiguration
        return self._config or GlobalConfiguration()

    def activate_stack(self, stack_name: str) -> None:
        """Set the active stack for the profile.
        Args:
            stack_name: name of the stack to activate
        """
        self.active_stack = stack_name
        self.global_config._write_config()

    class Config:
        """Pydantic configuration class."""

        # Validate attributes when assigning them. We need to set this in order
        # to have a mix of mutable and immutable attributes
        validate_assignment = True
        # Ignore extra attributes from configs of previous CoalescenceML versions
        extra = "ignore"
        # all attributes with leading underscore are private and therefore
        # are mutable and not included in serialization
        underscore_attrs_are_private = True



"""
## Pipelines

A CoalescenceML pipeline is a sequence of tasks that execute in a specific order and
yield artifacts. The artifacts are stored within the artifact store and indexed
via the metadata store. Each individual task within a pipeline is known as a
step.
"""


from coalescenceml.pipeline.base_pipeline import BasePipeline
from coalescenceml.pipeline.pipeline_decorator import pipeline
from coalescenceml.pipeline.schedule import Schedule


__all__ = ["BasePipeline", "pipeline", "Schedule"]



from abc import ABC, abstractmethod
from typing import ClassVar, List

from coalescenceml.enums import StackComponentFlavor
from coalescenceml.stack import StackComponent


class BaseStepOperator(StackComponent, ABC):
    """Base class for all CoalescenceML step operators."""

    # Class Configuration
    TYPE: ClassVar[StackComponentFlavor] = StackComponentFlavor.STEP_OPERATOR

    @abstractmethod
    def launch(
        self,
        pipeline_name: str,
        run_name: str,
        requirements: List[str],
        entrypoint_command: List[str],
    ) -> None:
        """Abstract method to execute a step.

        Concrete step operator subclasses must implement the following
        functionality in this method:
        - Prepare the execution environment and install all the necessary
          `requirements`
        - Launch a **synchronous** job that executes the `entrypoint_command`

        Args:
            pipeline_name: Name of the pipeline which the step to be executed
                is part of.
            run_name: Name of the pipeline run which the step to be executed
                is part of.
            entrypoint_command: Command that executes the step.
            requirements: List of pip requirements that must be installed
                inside the step operator environment.
        """



from collections import defaultdict
from typing import DefaultDict, Dict

from pydantic import BaseModel, validator

from coalescenceml.enums import StackComponentFlavor


class StackStoreModel(BaseModel):
    """Pydantic object used for serializing a CoalescenceML Stack Store.

    Attributes:
        version: coalescenceml version number
        stacks: Maps stack names to a configuration object containing the
            names and flavors of all stack components.
        stack_components: Contains names and flavors of all registered stack
            components.
    """

    stacks: Dict[str, Dict[StackComponentFlavor, str]]
    stack_components: DefaultDict[StackComponentFlavor, Dict[str, str]]

    @validator("stack_components")
    def _construct_defaultdict(
        cls, stack_components: Dict[StackComponentFlavor, Dict[str, str]]
    ) -> DefaultDict[StackComponentFlavor, Dict[str, str]]:
        """Ensures that `stack_components` is a defaultdict so stack
        components of a new component type can be added without issues."""
        return defaultdict(dict, stack_components)

    @classmethod
    def empty_store(cls) -> "StackStoreModel":
        """Initialize a new empty stack store with current zen version."""
        return cls(stacks={}, stack_components={})

    class Config:
        """Pydantic configuration class."""

        # Validate attributes when assigning them. We need to set this in order
        # to have a mix of mutable and immutable attributes
        validate_assignment = True
        # Ignore extra attributes from configs of previous CoalescenceML versions
        extra = "ignore"



from __future__ import annotations

import json
import logging
import time
import uuid
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    cast,
)

import tfx.orchestration.pipeline as tfx_pipeline
from pydantic import BaseModel
from tfx.orchestration.portable import data_types, launcher
from tfx.proto.orchestration.pipeline_pb2 import ContextSpec, PipelineNode

from coalescenceml.directory import Directory
from coalescenceml.logger import get_logger
from coalescenceml.step import BaseStep
from coalescenceml.step.utils import (
    INTERNAL_EXECUTION_PARAMETER_PREFIX,
    PARAM_PIPELINE_PARAMETER_NAME,
)
from coalescenceml.utils import readability_utils


if TYPE_CHECKING:
    from coalescenceml.pipeline.base_pipeline import BasePipeline
    from coalescenceml.pipeline.runtime_configuration import (
        RuntimeConfiguration,
    )
    from coalescenceml.stack import Stack

logger = get_logger(__name__)


def create_tfx_pipeline(
    coalescenceml_pipeline: BasePipeline, stack: Stack
) -> tfx_pipeline.Pipeline:
    """Creates a tfx pipeline from a CoalescenceML pipeline."""
    # Build dag from all steps in the pipeline
    coalescenceml_pipeline.build_dag(**coalescenceml_pipeline.steps)

    tfx_components = [
        step.component for step in coalescenceml_pipeline.steps.values()
    ]

    artifact_store = stack.artifact_store
    metadata_store = stack.metadata_store

    return tfx_pipeline.Pipeline(
        pipeline_name=coalescenceml_pipeline.name,
        components=tfx_components,
        pipeline_root=artifact_store.path,
        metadata_connection_config=metadata_store.get_tfx_metadata_config(),
        enable_cache=coalescenceml_pipeline.enable_cache,
    )


def get_cache_status(
    execution_info: data_types.ExecutionInfo,
) -> bool:
    """Returns the caching status of a step.

    Args:
        execution_info: The execution info of a `tfx` step.
    Raises:
        AttributeError: If the execution info is `None`.
        KeyError: If no pipeline info is found in the `execution_info`.
    Returns:
        The caching status of a `tfx` step as a boolean value.
    """
    if execution_info is None:
        logger.warning("No execution info found when checking cache status.")
        return False

    status = False
    directory = Directory()
    # TODO: Get the current running stack instead of just the active stack
    # It's possible to deploy a pipeline and switch stacks in which case we'll
    # run into issues....
    # However, the hope is most people will have similar stacks across profiles
    # So it'll inadvertently work LOL.
    active_stack = directory.active_stack
    if not active_stack:
        raise RuntimeError(
            "No active stack is configured for the directory. Run "
            "`coml stack set STACK_NAME` to update the active stack."
        )

    metadata_store = active_stack.metadata_store

    step_name_param = (
        INTERNAL_EXECUTION_PARAMETER_PREFIX + PARAM_PIPELINE_PARAMETER_NAME
    )
    step_name = json.loads(execution_info.exec_properties[step_name_param])
    if execution_info.pipeline_info:
        pipeline_name = execution_info.pipeline_info.id
    else:
        raise KeyError(f"No pipeline info found for step `{step_name}`.")
    pipeline_run_name = cast(str, execution_info.pipeline_run_id)
    pipeline = metadata_store.get_pipeline(pipeline_name)
    if pipeline is None:
        logger.error(f"Pipeline {pipeline_name} not found in Metadata Store.")
    else:
        status = (
            pipeline.get_run(pipeline_run_name).get_step(step_name).is_cached
        )
    return status


def execute_step(
    tfx_launcher: launcher.Launcher,
) -> Optional[data_types.ExecutionInfo]:
    """Executes a tfx component.
    Args:
        tfx_launcher: A tfx launcher to execute the component.
    Returns:
        Optional execution info returned by the launcher.
    """
    step_name_param = (
        INTERNAL_EXECUTION_PARAMETER_PREFIX + PARAM_PIPELINE_PARAMETER_NAME
    )
    pipeline_step_name = tfx_launcher._pipeline_node.node_info.id
    start_time = time.time()
    logger.info(f"Step `{pipeline_step_name}` has started.")
    try:
        execution_info = tfx_launcher.launch()
        if execution_info and get_cache_status(execution_info):
            if execution_info.exec_properties:
                step_name = json.loads(
                    execution_info.exec_properties[step_name_param]
                )
                logger.info(
                    f"Using cached version of `{pipeline_step_name}` "
                    f"[`{step_name}`].",
                )
            else:
                logger.error(
                    f"No execution properties found for step "
                    f"`{pipeline_step_name}`."
                )
    except RuntimeError as e:
        if "execution has already succeeded" in str(e):
            # Hacky workaround to catch the error that a pipeline run with
            # this name already exists. Raise an error with a more descriptive
            # message instead.
            raise RuntimeError(
                "Unable to run a pipeline with a run name that "
                "already exists."
            )
        else:
            raise e

    run_duration = time.time() - start_time
    logger.info(
        f"Step `{pipeline_step_name}` has finished in "
        f"{readability_utils.get_human_readable_time(run_duration)}."
    )
    return execution_info


def get_step_for_node(node: PipelineNode, steps: List[BaseStep]) -> BaseStep:
    """Finds the matching step for a tfx pipeline node."""
    step_name = node.node_info.id
    try:
        return next(step for step in steps if step.name == step_name)
    except StopIteration:
        raise RuntimeError(f"Unable to find step with name '{step_name}'.")


##### CONTEXT - RELATED UTILITIES #####
def add_context_to_node(
    pipeline_node: PipelineNode,
    type_: str,
    name: str,
    properties: Dict[str, str],
) -> None:
    """
    Add a new context to a TFX protobuf pipeline node.
    Args:
        pipeline_node: A tfx protobuf pipeline node
        type_: The type name for the context to be added
        name: Unique key for the context
        properties: dictionary of strings as properties of the context
    """
    # Add a new context to the pipeline
    context: ContextSpec = pipeline_node.contexts.contexts.add()
    # Adding the type of context
    context.type.name = type_
    # Setting the name of the context
    context.name.field_value.string_value = name
    # Setting the properties of the context depending on attribute type
    for key, value in properties.items():
        c_property = context.properties[key]
        c_property.field_value.string_value = value


def serialize_pydantic_object(
    obj: BaseModel, *, skip_errors: bool = False
) -> Dict[str, str]:
    """Convert a pydantic object to a dict of strings"""

    class PydanticEncoder(json.JSONEncoder):
        def default(self, o: Any) -> Any:
            try:
                return cast(Callable[[Any], str], obj.__json_encoder__)(o)
            except TypeError:
                return super().default(o)

    def _inner_generator(
        dictionary: Dict[str, Any]
    ) -> Iterator[Tuple[str, str]]:
        """Itemwise serialize each element in a dictionary."""
        for key, item in dictionary.items():
            try:
                yield key, json.dumps(item, cls=PydanticEncoder)
            except TypeError as e:
                if skip_errors:
                    logging.info(
                        "Skipping adding field '%s' to metadata context as "
                        "it cannot be serialized due to %s.",
                        key,
                        e,
                    )
                else:
                    raise TypeError(
                        f"Invalid type {type(item)} for key {key} can not be "
                        "serialized."
                    ) from e

    return {key: value for key, value in _inner_generator(obj.dict())}


def add_runtime_configuration_to_node(
    pipeline_node: PipelineNode,
    runtime_config: RuntimeConfiguration,
) -> None:
    """
    Add the runtime configuration of a pipeline run to a protobuf pipeline node.
    Args:
        pipeline_node: a tfx protobuf pipeline node
        runtime_config: a CoalescenceML RuntimeConfiguration
    """
    skip_errors: bool = runtime_config.get(
        "ignore_unserializable_fields", False
    )

    # Determine the name of the context
    def _name(obj: "BaseModel") -> str:
        """Compute a unique context name for a pydantic BaseModel."""
        try:
            return str(hash(obj.json(sort_keys=True)))
        except TypeError as e:
            class_name = obj.__class__.__name__
            logging.info(
                "Cannot convert %s to json, generating uuid instead. Error: %s",
                class_name,
                e,
            )
            return f"{class_name}_{uuid.uuid1()}"

    # iterate over all attributes of runtime context, serializing all pydantic
    # objects to node context.
    for key, obj in runtime_config.items():
        if isinstance(obj, BaseModel):
            logger.debug("Adding %s to context", key)
            add_context_to_node(
                pipeline_node,
                type_=obj.__repr_name__().lower(),
                name=_name(obj),
                properties=serialize_pydantic_object(
                    obj, skip_errors=skip_errors
                ),
            )



import functools
from typing import Any, Callable, Optional, Type, TypeVar, Union, cast, overload

from coalescenceml.directory import Directory
from coalescenceml.environment import Environment
from coalescenceml.integrations.mlflow.experiment_tracker import (
    MLFlowExperimentTracker,
)
from coalescenceml.logger import get_logger
from coalescenceml.step import BaseStep
from coalescenceml.step.utils import STEP_INNER_FUNC_NAME

logger = get_logger(__name__)


# Entry point function type
F = TypeVar("F", bound=Callable[..., Any])

# Step type
S = TypeVar("S", bound=Type[BaseStep])


@overload
def enable_mlflow(
    step: S,
) -> S:
    """Type annotation for no arguments."""
    ...


@overload
def enable_mlflow(
    experiment_name: str,
) -> Callable[[S], S]:
    """Type annotation for arguments."""
    ...


def enable_mlflow(
    step: S,
    experiment_name: Optional[str] = None,
) -> Union[S, Callable[[S], S]]:
    """Decorator to enable MLFlow run for a step.

    Args:
        step: Step class to be decorated
        experiment_name: Name of mlflow experiment to use. Defaults
            to pipeline name

    Returns:
        A BaseStep with wrapped entrypoint functionality
    """
    def inner_decorator(step_: S) -> S:
        """Decorate step entrypoint."""

        logger.debug(
            f"Applying MLFlow decorator to step {step_.__name__}"
        )
        if not issubclass(step_, BaseStep):
            raise RuntimeError(
                "The MLFlow decorator can only be applied to a BaseStep "
                "subclass or a function decorated with @step from coml."
            )

        entrypoint_orig = getattr(step_, STEP_INNER_FUNC_NAME)
        return cast(
            S,
            type(
                step_.__name__,
                (step_,),
                {
                    STEP_INNER_FUNC_NAME: staticmethod(
                        mlflow_entrypoint(
                            experiment_name=experiment_name,
                        )(entrypoint_orig)
                    ),
                    "__module__": step_.__module__,
                }
            )
        )

    if step is None:
        return inner_decorator
    else:
        return inner_decorator(step)


def mlflow_entrypoint(
    experiment_name: Optional[str] = None,
) -> Callable[[F], F]:
    """Build entrypoint for wandb experiment.

    Args:
        experiment_name:

    Returns:
        wrapped function with mlflow access
    """
    # TODO: Modify the experiment tracker to consider the
    # project name as well as experiment name being passed
    # in explicitly to the decorator.
    def inner_decorator(func: F) -> F:
        """Wrap function with wandb experiment tracker

        Args:
            func: function to wrap

        Returns:
            Wrapped function
        """
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            """Wrap function.

            Args:
                *args: pass through
                **kwargs: pass through

            Returns:
                passthrough
            """
            experiment_tracker = Directory(
                skip_directory_check=True,
            ).active_stack.experiment_tracker

            if not isinstance(experiment_tracker, MLFlowExperimentTracker):
                raise ValueError(
                    "The active stack must have an MLFlow experiment tracker "
                    "to be registered to be able to track experiments using "
                    "MLFlow."
            )

            active_run = experiment_tracker.active_run(experiment_name=experiment_name)
            if not active_run:
                raise RuntimeError(
                    "No active MLFlow run configured."
                )

            with active_run:
                return func(*args, **kwargs)
        
        return cast(F, wrapper)

    return inner_decorator




import shutil
from typing import Any, Dict, List, Tuple, Type, cast

import pkg_resources

from coalescenceml.integrations.registry import integration_registry
from coalescenceml.logger import get_logger


logger = get_logger(__name__)


class IntegrationMeta(type):
    """Metaclass responsible for registering different Integration
    subclasses"""

    def __new__(
        mcs, name: str, bases: Tuple[Type[Any], ...], dct: Dict[str, Any]
    ) -> "IntegrationMeta":
        """Hook into creation of an Integration class."""
        cls = cast(Type["Integration"], super().__new__(mcs, name, bases, dct))
        if name != "Integration":
            integration_registry.register_integration(cls.NAME, cls)
        return cls


class Integration(metaclass=IntegrationMeta):
    """Base class for integration in CoalescenceML"""

    NAME = "base_integration"

    REQUIREMENTS: List[str] = []

    # tool name : command to check
    SYSTEM_REQUIREMENTS: Dict[str, str] = {}

    @classmethod
    def check_installation(cls) -> bool:
        """Method to check whether the required packages are installed"""
        try:
            for requirement, command in cls.SYSTEM_REQUIREMENTS.items():
                result = shutil.which(command)

                if result is None:
                    logger.debug(
                        "Unable to find the required packages for %s on your "
                        "system. Please install the packages on your system "
                        "and try again.",
                        requirement,
                    )
                    return False

            for r in cls.REQUIREMENTS:
                pkg_resources.get_distribution(r)
            logger.debug(
                f"Integration {cls.NAME} is installed correctly with "
                f"requirements {cls.REQUIREMENTS}."
            )
            return True
        except pkg_resources.DistributionNotFound as e:
            logger.debug(
                f"Unable to find required package '{e.req}' for "
                f"integration {cls.NAME}."
            )
            return False
        except pkg_resources.VersionConflict as e:
            logger.debug(
                f"VersionConflict error when loading installation {cls.NAME}: "
                f"{str(e)}"
            )
            return False

    @staticmethod
    def activate() -> None:
        """Abstract method to activate the integration"""



"""
The CoalescenceML integrations module contains sub-modules for each integration that we
support. This includes orchestrators like Apache Airflow as well as deep
learning libraries like PyTorch.
"""
# Imports here when we have them lol
from coalescenceml.integrations.mlflow import MLFlowIntegration
from coalescenceml.integrations.azure_datalake import AzureIntegration
from coalescenceml.integrations.sklearn import SKLearnIntegration
from coalescenceml.integrations.statsmodels import StatsmodelsIntegration
from coalescenceml.integrations.tensorflow import TFIntegration
from coalescenceml.integrations.xgboost import XgboostIntegration


from coalescenceml.producers.base_producer import BaseProducer
from coalescenceml.producers.json_producer import JSONProducer
from coalescenceml.producers.numpy_producer import NumpyProducer


__all__ = [
    "BaseProducer",
    "JSONProducer",
    "NumpyProducer",
]



import hashlib
import importlib
import inspect
import os
import pathlib
import site
import sys
from contextlib import contextmanager
from types import (
    CodeType,
    FrameType,
    FunctionType,
    MethodType,
    ModuleType,
    TracebackType,
)
from typing import Any, Callable, Iterator, Optional, Type, Union

from coalescenceml.environment import Environment
from coalescenceml.logger import get_logger


logger = get_logger(__name__)


def get_source(value: Any) -> str:
    """Returns the source code of an object. If executing within a IPython
    kernel environment, then this monkey-patches `inspect` module temporarily
    with a workaround to get source from the cell.
    Raises:
        TypeError: If source not found.
    """
    if Environment.in_notebook():
        # Hacky patch inspect.getfile temporarily to make get_source work.
        # Source: https://stackoverflow.com/questions/51566497/
        # The BS that jupyter notebooks make us do...
        def _new_getfile(
            object: Any,
            _old_getfile: Callable[
                [
                    Union[
                        ModuleType,
                        Type[Any],
                        MethodType,
                        FunctionType,
                        TracebackType,
                        FrameType,
                        CodeType,
                        Callable[..., Any],
                    ]
                ],
                str,
            ] = inspect.getfile,
        ) -> Any:
            if not inspect.isclass(object):
                return _old_getfile(object)

            # Lookup by parent module (as in current inspect)
            if hasattr(object, "__module__"):
                object_ = sys.modules.get(object.__module__)
                if hasattr(object_, "__file__"):
                    return object_.__file__

            # If parent module is __main__, lookup by methods
            for name, member in inspect.getmembers(object):
                if (
                    inspect.isfunction(member)
                    and object.__qualname__ + "." + member.__name__
                    == member.__qualname__
                ):
                    return inspect.getfile(member)
            else:
                raise TypeError(f"Source for {object!r} not found.")

        # Hacky patch, compute source, then revert Hacky patch.
        _old_getfile = inspect.getfile
        inspect.getfile = _new_getfile
        try:
            src = inspect.getsource(value)
        finally:
            inspect.getfile = _old_getfile
    else:
        # Use standard inspect if running outside a notebook
        src = inspect.getsource(value)
    return src


def get_hashed_source(value: Any) -> str:
    """Returns a hash of the objects source code."""
    try:
        source_code = get_source(value)
    except TypeError:
        raise TypeError(
            f"Unable to compute the hash of source code of object: {value}."
        )
    return hashlib.sha256(source_code.encode("utf-8")).hexdigest()


@contextmanager
def prepend_python_path(path: str) -> Iterator[None]:
    """Simple context manager to help import module within the repo"""
    try:
        # Entering the with statement
        sys.path.insert(0, path)
        yield
    finally:
        # Exiting the with statement
        sys.path.remove(path)


def import_class_by_path(class_path: str) -> Type[Any]:
    """Imports a class based on a given path

    Args:
        class_path: str, class_source e.g. this.module.Class
    Returns: the given class
    """
    classname = class_path.split(".")[-1]
    modulename = ".".join(class_path.split(".")[0:-1])
    mod = importlib.import_module(modulename)
    return getattr(mod, classname)


def load_source_path_class(
    source: str, import_path: Optional[str] = None
) -> Type[Any]:
    """Loads a Python class from the source.
    Args:
        source: class_source e.g. this.module.Class[@sha]
        import_path: optional path to add to python path
    """
    from coalescenceml.directory import Directory

    repo_root = Directory.find_directory()
    if not import_path and repo_root:
        import_path = str(repo_root)

    if "@" in source:
        source = source.split("@")[0]

    if import_path is not None:
        with prepend_python_path(import_path):
            logger.debug(
                f"Loading class {source} with import path {import_path}"
            )
            return import_class_by_path(source)
    return import_class_by_path(source)


def get_source_root_path() -> str:
    """Get the directory root path or the source root path of the current
    process.

    Returns:
        The source root path of the current process.
    """
    from coalescenceml.directory import Directory

    dir_root = Directory.find_directory()
    if dir_root:
        logger.debug("Using directory root as source root: %s", dir_root)
        return str(dir_root.resolve())

    main_module = sys.modules.get("__main__")
    if main_module is None:
        raise RuntimeError(
            "Could not determine the main module used to run the current "
            "process."
        )

    if not hasattr(main_module, "__file__") or not main_module.__file__:
        raise RuntimeError(
            "Main module was not started from a file. Cannot "
            "determine the module root path."
        )
    path = pathlib.Path(main_module.__file__).resolve().parent

    logger.debug("Using main module location as source root: %s", path)
    return str(path)


def get_module_source_from_module(module: ModuleType) -> str:
    """Gets the source of the supplied module.

    Args:
        module: the module to get the source of.
    Returns:
        The source of the main module.
    Raises:
        RuntimeError: if the module is not loaded from a file
    """
    if not hasattr(module, "__file__") or not module.__file__:
        if module.__name__ == "__main__":
            raise RuntimeError(
                f"{module} module was not loaded from a file. Cannot "
                "determine the module root path."
            )
        return module.__name__
    module_path = os.path.abspath(module.__file__)

    root_path = get_source_root_path()

    if not module_path.startswith(root_path):
        root_path = os.getcwd()
        logger.warning(
            "User module %s is not in the source root. Using current "
            "directory %s instead to resolve module source.",
            module,
            root_path,
        )

    # Remove root_path from module_path to get relative path left over
    module_path = module_path.replace(root_path, "")[1:]

    # Kick out the .py and replace `/` with `.` to get the module source
    module_path = module_path.replace(".py", "")
    module_source = module_path.replace("/", ".")

    logger.debug(
        f"Resolved module source for module {module} to: {module_source}"
    )

    return module_source


def is_standard_source(source: str) -> bool:
    """Returns `True` if source is a standard CoalescenceML source.

    Args:
        source: class_source e.g. this.module.Class[@pin].
    """
    if source.split(".")[0] == "coalescenceml":
        return True
    return False


def is_third_party_module(file_path: str) -> bool:
    """Returns whether a file belongs to a third party package."""
    absolute_file_path = pathlib.Path(file_path).resolve()

    for path in site.getsitepackages() + [site.getusersitepackages()]:
        if pathlib.Path(path).resolve() in absolute_file_path.parents:
            return True

    return False


def resolve_class(class_: Type[Any]) -> str:
    """Resolves a class into a serializable source string.
    For classes that are not built-in nor imported from a Python package, the
    `get_source_root_path` function is used to determine the root path
    relative to which the class source is resolved.
    Args:
        class_: A Python Class reference.
    Returns: source_path e.g. this.module.Class.
    """
    initial_source = class_.__module__ + "." + class_.__name__
    if is_standard_source(initial_source):
        return initial_source

    try:
        file_path = inspect.getfile(class_)
    except TypeError:
        # builtin file
        return initial_source

    if initial_source.startswith("__main__") or is_third_party_module(
        file_path
    ):
        return initial_source

    # Regular user file -> get the full module path relative to the
    # source root.
    module_source = get_module_source_from_module(
        sys.modules[class_.__module__]
    )

    # ENG-123 Sanitize for Windows OS
    # module_source = module_source.replace("\\", ".")

    logger.debug(f"Resolved class {class_} to {module_source}")

    return module_source + "." + class_.__name__



AIRFLOW = "airflow"  # Orchestrator
AWS = "aws"  # Key vault / Secrets
AWS_ENDPOINT_S3 = "https://s3." # AWS Artifact Store
AWS_ENDPOINT_URL = ".amazonaws.com" # AWS Artifact Store
AZURE = "azure"  # Artifact store, Key vault
AZUREML = "azureml"  # Step controller (training)
EVIDENTLY = "evidently"  # Profiling
GCP = "gcp"  # Artifact store
KUBEFLOW = "kubeflow"  # Orchestrator, Metadata store
MLFLOW = "mlflow"  # Deployer, logging, context
PYTORCH = "pytorch"  # Special producers
PYTORCH_L = "pytorch_lightning"  # Special producers
S3 = "s3"  # Artifact store
SAGEMAKER = "sagemaker"  # Step controller (training / inference)
SELDON = "seldon"  # Deployer (with Alibi we can do monitoring too!)
SKLEARN = "sklearn"  # Special train steps; Producers
STATSMODELS = "statsmodels" # Producers;
TENSORFLOW = "tensorflow"  # Special train step; Producers
WHYLOGS = "whylogs"  # Data / Model profiling
VERTEX = "vertex"  # Orchestrator, Step controller
XGBOOST = "xgboost" # producers



"""
A step is a single piece or stage of an ML Pipeline. Each step is essentilly
a node of a Directed Acyclic Graph (or DAG). 

Steps can be subclassed from the `BaseStep` class, or used via our `@step` decorator.
"""

from coalescenceml.step.base_step import BaseStep
from coalescenceml.step.base_step_config import BaseStepConfig
from coalescenceml.step.output import Output
from coalescenceml.step.step_context import StepContext
from coalescenceml.step.step_decorator import step
from coalescenceml.step.step_environment import (
    STEP_ENVIRONMENT_NAME,
    StepEnvironment,
)


__all__ = [
    "STEP_ENVIRONMENT_NAME",
    "BaseStep",
    "BaseStepConfig",
    "StepContext",
    "step",
    "Output",
    "StepEnvironment",
]



import os
from typing import Any, Type

import tensorflow as tf

from coalescenceml.artifacts import DataArtifact
from coalescenceml.io import fileio
from coalescenceml.logger import get_logger
from coalescenceml.producers.base_producer import BaseProducer
from coalescenceml.producers.producer_registry import register_producer_class


logger = get_logger(__name__)
DEFAULT_FILENAME = "tf_saved_data"


@register_producer_class
class TensorflowDatasetProducer(BaseProducer):
    """Read/Write TF Dataset files."""

    ARTIFACT_TYPES = (DataArtifact,)
    TYPES = (tf.data.Dataset,)

    def handle_input(self, data_type: Type[Any]) -> tf.data.Dataset:
        """Read tf.data.Dataset data into memory."""
        super().handle_input(data_type)
        filepath = os.path.join(self.artifact.uri, DEFAULT_FILENAME)
        contents = tf.data.experimental.load(filepath)
        return contents

    def handle_return(self, dataset: tf.data.Dataset) -> None:
        """Writes a tf.data.Dataset object to filesystem."""
        super().handle_return(model)
        filepath = os.path.join(self.artifact.uri, DEFAULT_FILENAME)
        tf.data.experimental.save(
            dataset, filepath, compression=None, shard_func=None,
        )



from coalescenceml.integrations.integration import Integration

class MLFlowIntegration(Integration):
    NAME = "mlflow_integration"
    REQUIREMENTS = ["mlflow==1.25"]
    SYSTEM_REQUIREMENTS = { "docker-engine" : "docker", "kubectl" : "kubectl" }


from coalescenceml.artifacts.data_analysis_artifact import DataAnalysisArtifact
from coalescenceml.artifacts.data_artifact import DataArtifact
from coalescenceml.artifacts.metrics_artifact import MetricsArtifact
from coalescenceml.artifacts.model_artifact import ModelArtifact


__all__ = [
    "DataAnalysisArtifact",
    "DataArtifact",
    "ModelArtifact",
    "MetricsArtifact",
]



import glob
import os
import shutil
from typing import (
    Any,
    Callable,
    ClassVar,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)

from pydantic import validator

from coalescenceml.artifact_store import BaseArtifactStore
from coalescenceml.artifact_store.exceptions import ArtifactStoreInterfaceError


# TODO: Can we have this be imported from somewhere b/c we use it a lot??
PathType = Union[bytes, str]


class LocalArtifactStore(BaseArtifactStore):
    """Artifact Store for local artifacts."""

    # Class Configuration
    FLAVOR: ClassVar[str] = "local"
    SUPPORTED_SCHEMES: ClassVar[Set[str]] = {"", "file://"}

    @staticmethod
    def open(name: PathType, mode: str = "r") -> Any:
        """Open a file at the given path."""
        return open(name, mode=mode)

    @staticmethod
    def copyfile(src: PathType, dst: PathType, overwrite: bool = False) -> None:
        """Copy a file from the source to the destination."""
        if not overwrite and os.path.exists(dst):
            raise FileExistsError(
                f"Destination file {str(dst)} already exists and argument "
                f"`overwrite` is false."
            )
        shutil.copyfile(src, dst)

    @staticmethod
    def exists(path: PathType) -> bool:
        """Returns `True` if the given path exists."""
        return os.path.exists(path)

    @staticmethod
    def glob(pattern: PathType) -> List[PathType]:
        """Return the paths that match a glob pattern."""
        return glob.glob(pattern)

    @staticmethod
    def isdir(path: PathType) -> bool:
        """Returns whether the given path points to a directory."""
        return os.path.isdir(path)

    @staticmethod
    def listdir(path: PathType) -> List[PathType]:
        """Returns a list of files under a given directory in the filesystem."""
        return os.listdir(path)

    @staticmethod
    def makedirs(path: PathType) -> None:
        """Make a directory at the given path, recursively creating parents."""
        os.makedirs(path, exist_ok=True)

    @staticmethod
    def mkdir(path: PathType) -> None:
        """Make a directory at the given path; parent directory must exist."""
        os.mkdir(path)

    @staticmethod
    def remove(path: PathType) -> None:
        """Remove the file at the given path. Dangerous operation."""
        os.remove(path)

    @staticmethod
    def rename(src: PathType, dst: PathType, overwrite: bool = False) -> None:
        """Rename source file to destination file.
        Args:
            src: The path of the file to rename.
            dst: The path to rename the source file to.
            overwrite: If a file already exists at the destination, this
                method will overwrite it if overwrite=`True`
        """
        if not overwrite and os.path.exists(dst):
            raise FileExistsError(
                f"Destination path {str(dst)} already exists and argument "
                f"`overwrite` is false."
            )
        os.rename(src, dst)

    @staticmethod
    def rmtree(path: PathType) -> None:
        """Deletes dir recursively. Dangerous operation."""
        shutil.rmtree(path)

    @staticmethod
    def stat(path: PathType) -> Any:
        """Return the stat descriptor for a given file path."""
        return os.stat(path)

    @staticmethod
    def walk(
        top: PathType,
        topdown: bool = True,
        onerror: Optional[Callable[..., None]] = None,
    ) -> Iterable[Tuple[PathType, List[PathType], List[PathType]]]:
        """Return an iterator that walks the contents of the given directory.
        Args:
            top: Path of directory to walk.
            topdown: Whether to walk directories topdown or bottom-up.
            onerror: Callable that gets called if an error occurs.
        Returns:
            An Iterable of Tuples, each of which contain the path of the
            current directory path, a list of directories inside the
            current directory and a list of files inside the current
            directory.
        """
        yield from os.walk(top, topdown=topdown, onerror=onerror)

    @validator("path")
    def ensure_path_local(cls, path: str) -> str:
        from coalescenceml.constants import REMOTE_FS_PREFIX

        if any(path.startswith(prefix) for prefix in REMOTE_FS_PREFIX):
            raise ArtifactStoreInterfaceError(
                f"The path:{path} you defined for your local artifact store "
                f"start with one of the remote prefixes."
            )
        return path



"""
## Metadata Store
The configuration of each pipeline, step, backend, and produced artifacts are
all tracked within the metadata store. The metadata store is an SQL database,
and can be `sqlite` or `mysql`.

Metadata are the pieces of information tracked about the pipelines, experiments
and configurations that you are running with CoML. 
"""

from coalescenceml.metadata_store.base_metadata_store import BaseMetadataStore
from coalescenceml.metadata_store.mysql_metadata_store import MySQLMetadataStore
from coalescenceml.metadata_store.sqlite_metadata_store import (
    SQLiteMetadataStore,
)


__all__ = [
    "BaseMetadataStore",
    "MySQLMetadataStore",
    "SQLiteMetadataStore",
]



"""
## Artifact Stores

the inputs and outputs which go through any step is treated as an
artifact and as its name suggests, an `ArtifactStore` is a place where these
artifacts get stored.
"""
from coalescenceml.artifact_store.base_artifact_store import BaseArtifactStore
from coalescenceml.artifact_store.local_artifact_store import LocalArtifactStore


__all__ = ["BaseArtifactStore", "LocalArtifactStore"]



import base64
import json
from uuid import UUID

import yaml
from pydantic import BaseModel

from coalescenceml.enums import StackComponentFlavor
from coalescenceml.stack import StackComponent


class StackComponentWrapper(BaseModel):
    """Serializable Configuration of a StackComponent"""

    type: StackComponentFlavor
    flavor: str  # due to subclassing, can't properly use enum type here
    name: str
    uuid: UUID
    config: bytes  # b64 encoded yaml config

    @classmethod
    def from_component(
        cls, component: StackComponent
    ) -> "StackComponentWrapper":
        return cls(
            type=component.TYPE,
            flavor=component.FLAVOR,
            name=component.name,
            uuid=component.uuid,
            config=base64.b64encode(
                yaml.dump(json.loads(component.json())).encode()
            ),
        )



from coalescenceml.stack_store.model.stack_component_wrapper import (
    StackComponentWrapper,
)
from coalescenceml.stack_store.model.stack_store_model import StackStoreModel
from coalescenceml.stack_store.model.stack_wrapper import StackWrapper


__all__ = [
    "StackComponentWrapper",
    "StackStoreModel",
    "StackWrapper",
]



from coalescenceml.integrations.constants import STATSMODELS
from coalescenceml.integrations.integration import Integration


class StatsmodelsIntegration(Integration):
    """Integration for statsmodels."""

    NAME = STATSMODELS
    REQUIREMENTS = ["statsmodels==0.13.2"]

    @classmethod
    def activate(cls) -> None:
        from coalescenceml.integrations.statsmodels import producers


StatsmodelsIntegration.check_installation()



def get_human_readable_time(seconds: float) -> str:
    """Convert seconds into a human-readable string.

    Args:
        seconds: number of seconds

    Returns:
        String that displays time in Days Hours, Minutes, Seconds format
    """
    prefix = "-" if seconds < 0 else ""
    seconds = abs(seconds)
    int_seconds = int(seconds)
    days, int_seconds = divmod(int_seconds, 86400)
    hours, int_seconds = divmod(int_seconds, 3600)
    minutes, int_seconds = divmod(int_seconds, 60)
    if days > 0:
        time_string = f"{days}d{hours}h{minutes}m{int_seconds}s"
    elif hours > 0:
        time_string = f"{hours}h{minutes}m{int_seconds}s"
    elif minutes > 0:
        time_string = f"{minutes}m{int_seconds}s"
    else:
        time_string = f"{seconds:.3f}s"

    return prefix + time_string



from __future__ import annotations

from typing import TYPE_CHECKING, Any, List

from coalescenceml.logger import get_logger


if TYPE_CHECKING:
    from coalescenceml.metadata_store import BaseMetadataStore
    from coalescenceml.post_execution.pipeline_run import PipelineRunView

logger = get_logger(__name__)


class PipelineView:
    """Post-execution pipeline class which can be used to query
    pipeline-related information from the metadata store.
    """

    def __init__(self, id_: int, name: str, metadata_store: BaseMetadataStore):
        """Initializes a post-execution pipeline object.
        In most cases `PipelineView` objects should not be created manually
        but retrieved using the `get_pipelines()` method of a
        `coalescenceml.directory.Directory` instead.
        Args:
            id_: The context id of this pipeline.
            name: The name of this pipeline.
            metadata_store: The metadata store which should be used to fetch
                additional information related to this pipeline.
        """
        self._id = id_
        self._name = name
        self._metadata_store = metadata_store

    @property
    def name(self) -> str:
        """Returns the name of the pipeline."""
        return self._name

    @property
    def runs(self) -> List[PipelineRunView]:
        """Returns all stored runs of this pipeline.
        The runs are returned in chronological order, so the latest
        run will be the last element in this list.
        """
        # Do not cache runs as new runs might appear during this objects
        # lifecycle
        runs = self._metadata_store.get_pipeline_runs(self)
        return list(runs.values())

    def get_run_names(self) -> List[str]:
        """Returns a list of all run names."""
        # Do not cache runs as new runs might appear during this objects
        # lifecycle
        runs = self._metadata_store.get_pipeline_runs(self)
        return list(runs.keys())

    def get_run(self, name: str) -> "PipelineRunView":
        """Returns a run for the given name.
        Args:
            name: The name of the run to return.
        Raises:
            KeyError: If there is no run with the given name.
        """
        run = self._metadata_store.get_pipeline_run(self, name)
        if not run:
            raise KeyError(
                f"No run found for name `{name}`. This pipeline "
                f"only has runs with the following "
                f"names: `{self.get_run_names()}`"
            )
        return run

    def get_run_for_completed_step(self, step_name: str) -> PipelineRunView:
        """This method helps you find out which pipeline run produced
        the cached artifact of a given step.
        Args:
            step_name: Name of step at hand
        Return:
            None if no run is found that completed the given step,
             else the original pipeline_run
        """
        orig_pipeline_run = None

        for run in reversed(self.runs):
            try:
                step = run.get_step(step_name)
                if step.is_completed:
                    orig_pipeline_run = run
                    break
            except KeyError:
                pass
        if not orig_pipeline_run:
            raise LookupError(
                "No Pipeline Run could be found, that has"
                f" completed the provided step: [{step_name}]"
            )

        return orig_pipeline_run

    def __repr__(self) -> str:
        """Returns a string representation of this pipeline."""
        return (
            f"{self.__class__.__qualname__}(id={self._id}, "
            f"name='{self._name}')"
        )

    def __eq__(self, other: Any) -> bool:
        """Returns whether the other object is referring to the
        same pipeline."""
        if isinstance(other, PipelineView):
            return (
                self._id == other._id
                and self._metadata_store.uuid == other._metadata_store.uuid
            )
        return False



"""
## Orchestrator

An orchestrator is a special kind of backend that manages the running of each
step of the pipeline. Orchestrators administer the actual pipeline runs.
"""
from coalescenceml.orchestrator.base_orchestrator import BaseOrchestrator
from coalescenceml.orchestrator.local_orchestrator import LocalOrchestrator


__all__ = ["BaseOrchestrator", "LocalOrchestrator"]



from typing import ClassVar, Union

from ml_metadata.proto import metadata_store_pb2
from pydantic import validator
from tfx.orchestration import metadata

from coalescenceml.io import utils
from coalescenceml.metadata_store import BaseMetadataStore


class SQLiteMetadataStore(BaseMetadataStore):
    """SQLite backend for CoalescenceML metadata store."""

    uri: str

    # Class Configuration
    FLAVOR: ClassVar[str] = "sqlite"

    def get_tfx_metadata_config(
        self,
    ) -> Union[
        metadata_store_pb2.ConnectionConfig,
        metadata_store_pb2.MetadataStoreClientConfig,
    ]:
        """Return tfx metadata config for sqlite metadata store."""
        return metadata.sqlite_metadata_connection_config(self.uri)

    @validator("uri")
    def ensure_uri_is_local(cls, uri: str) -> str:
        """Ensures that the metadata store uri is local."""
        if utils.is_remote(uri):
            raise ValueError(
                f"Uri '{uri}' specified for SQLiteMetadataStore is not a "
                f"local uri."
            )

        return uri



class PipelineInterfaceError(Exception):
    """Raises exception when interacting with the Pipeline interface
    in an unsupported way."""


class PipelineConfigurationError(Exception):
    """Raises exceptions when a pipeline configuration contains
    invalid values."""



class InitializationException(Exception):
    """Raised when an error occurred during init of a CoML directory."""


class ForbiddenDirectoryAccessError(RuntimeError):
    """Raised when accessing a CoML directory while a step is executed."""



"""
After executing a pipeline, the user needs to be able to fetch it from history
to interact with it. The post_execution submodule provides a set of
interfaces with which the user can interact with artifacts, the pipeline, steps,
and the post-run pipeline object.
"""

from coalescenceml.post_execution.artifact import ArtifactView
from coalescenceml.post_execution.pipeline import PipelineView
from coalescenceml.post_execution.pipeline_run import PipelineRunView
from coalescenceml.post_execution.step import StepView


__all__ = ["PipelineView", "PipelineRunView", "StepView", "ArtifactView"]



from abc import ABC, abstractmethod
from typing import Any, ClassVar

from coalescenceml.enums import StackComponentFlavor
from coalescenceml.pipeline import BasePipeline
from coalescenceml.pipeline.runtime_configuration import RuntimeConfiguration
from coalescenceml.stack import Stack, StackComponent


class BaseOrchestrator(StackComponent, ABC):
    """Base class for all CoalescenceML orchestrators."""

    # Class Configuration
    TYPE: ClassVar[StackComponentFlavor] = StackComponentFlavor.ORCHESTRATOR

    @abstractmethod
    def run_pipeline(
        self,
        pipeline: BasePipeline,
        stack: Stack,
        runtime_configuration: RuntimeConfiguration,
    ) -> Any:
        """Runs a pipeline.
        Args:
            pipeline: The pipeline to run.
            stack: The stack on which the pipeline is run.
            runtime_configuration: Runtime configuration of the pipeline run.
        """



import os
from typing import Optional

from rich.console import Console
from rich.theme import Theme


coml_custom_theme = Theme(
    {
        "info": "dim cyan",
        "warning": "yellow",
        "danger": "bold red",
        "error": "red",
        "title": "bold cyan underline",
    }
)

console = Console(theme=coml_custom_theme, markup=True)


def process_bool_env_var(var: str, default: bool = False) -> bool:
    """process_bool_env_car converts env var to boolean.

    Args:
        var: the variable to grab from the environment
        default: value to return if var does not exist

    Returns:
        boolean value to assign the variable
    """
    value = os.getenv(var)
    if value in {"1", "True", "true"}:
        return True
    else:
        return default


# Global
APP_NAME = "CoalescenceML"

# Env variables
ENV_COML_DEBUG = "COML_DEBUG"
ENV_COML_LOGGING_VERBOSITY = "COML_LOGGING_VERBOSITY"
ENV_COML_CONFIG_PATH = "COML_CONFIG_PATH"
ENV_COML_DEFAULT_STORE_TYPE = "COML_DEFAULT_STORE_TYPE"
ENV_COML_PREVENT_PIPELINE_EXECUTION = "COML_PREVENT_PIPELINE_EXECUTION"
ENV_COML_PROFILE_NAME = "COML_PROFILE_NAME"
ENV_COML_DIRECTORY_PATH = "COML_DIRECTORY_PATH"

# Logging variables
IS_DEBUG_ENV: bool = process_bool_env_var(ENV_COML_DEBUG, default=False)

COML_LOGGING_VERBOSITY: str = "INFO"

if IS_DEBUG_ENV:
    COML_LOGGING_VERBOSITY = os.getenv(
        ENV_COML_LOGGING_VERBOSITY, default="DEBUG"
    ).upper()
else:
    COML_LOGGING_VERBOSITY = os.getenv(
        ENV_COML_LOGGING_VERBOSITY, default="INFO"
    ).upper()

REMOTE_FS_PREFIX = ["gs://", "s3://", "az://"]


# Container utils
SHOULD_PREVENT_PIPELINE_EXECUTION = process_bool_env_var(
    ENV_COML_PREVENT_PIPELINE_EXECUTION
)

# DIRECTORY and local store directory paths:
DIRECTORY_DIRECTORY_NAME = ".coalescence"
LOCAL_STORES_DIRECTORY_NAME = "local_stores"

# SPECIAL KEYS
RUN_NAME_OPTION_KEY = "run_name"
USER_MAIN_MODULE: Optional[str] = None



"""
https://stackoverflow.com/questions/31875/is-there-a-simple-elegant-way-to-define-singletons
https://www.geeksforgeeks.org/singleton-pattern-in-python-a-complete-guide/
https://www.geeksforgeeks.org/singleton-method-python-design-patterns/

"""

from typing import Any, Optional, cast


class SingletonMetaClass(type):
    """Singleton metaclass.
    Use this metaclass to make any class into a singleton class:
    ```python
    class BorgSingleton(metaclass=SingletonMetaClass):
        def __init__(self, shared_var):
            self._shared_var = shared_var

        @property
        def shared_var(self):
            return self._shared_var
    first_instance = BorgSingleton('Algorithm')
    second_instance = BorgSingleton('Data Structure')
    print(second_instance.owner)  # Algorithm
    BorgSingleton._clear() # ring destroyed
    ```
    """

    def __init__(cls, *args: Any, **kwargs: Any) -> None:
        """Initialize a singleton class."""
        super().__init__(*args, **kwargs)
        cls.__singleton_instance: Optional["SingletonMetaClass"] = None

    def __call__(cls, *args: Any, **kwargs: Any) -> "SingletonMetaClass":
        """Create or return the singleton instance."""
        if not cls.__singleton_instance:
            cls.__singleton_instance = cast(
                "SingletonMetaClass", super().__call__(*args, **kwargs)
            )

        return cls.__singleton_instance

    def _clear(cls) -> None:
        """Clear the singleton instance."""
        cls.__singleton_instance = None



from __future__ import annotations

from inspect import isclass
from typing import Any, Dict, Type, TypeVar

# if TYPE_CHECKING:
from coalescenceml.artifacts.base_artifact import BaseArtifact
from coalescenceml.artifacts.type_registry import type_registry

# Exception
from coalescenceml.logger import get_logger
from coalescenceml.producers.base_producer import BaseProducer
from coalescenceml.producers.exceptions import ProducerInterfaceError


logger = get_logger(__name__)


class ProducerRegistry(object):
    """Matches a type to a default producer."""

    def __init__(self) -> None:
        self.producer_types: Dict[Type[Any], Type[BaseProducer]] = {}

    def register_producer(
        self, object_type: Type[Any], producer: BaseProducer
    ) -> None:
        """Registers a new producer.

        Args:
            object_type: Indicates python type of object.
            producer: A BaseProducer subclass for the python type.
        """
        if object_type in self.producer_types:
            logger.debug(
                f"Found existing producer for {object_type}: {self.producer_types[object_type]}. Skipping registration of {producer}."
            )
        else:
            self.producer_types[object_type] = producer
            logger.debug(f"Registered producer {producer} for {object_type}.")

    def register_overwrite_producer(
        self, object_type: Type[Any], producer: BaseProducer
    ) -> None:
        """Registers a new producer and overwrites any prior setting.

        Args:
            object_type: Indicates python type of object.
            producer: A BaseProducer subclass for the python type.
        """
        self.producer_types[object_type] = producer
        logger.debug(f"Registered producer {producer} for {object_type}.")

    def get_producers(self) -> Dict[Type[Any], Type[BaseProducer]]:
        """Get all registered producers."""
        return self.producer_types

    def is_registered(self, object_type: Type[Any]) -> bool:
        """Returns if a producer is registered for the given type.

        Args:
            object_type: Indicates python type of object.

        Returns:
            bool: Whether a producer class is registered for the python type
        """
        return any(issubclass(object_type, p) for p in self.producer_types)

    def __getitem__(self, object_type: Type[Any]) -> BaseProducer:
        """Retrieve a single producer based on the python type.

        Args:
            object_type: Indicates python type of object.

        Returns:
            `BaseProducer` subclass that was registered for this type.

        Raises:
            ValueError: If the type (or any of its superclasses) is not registered or the type has more than one superclass with different producers.
        """
        if object_type in self.producer_types:
            return self.producer_types[object_type]
        else:
            # if type is not registered then check for superclasses
            producers_compatible_superclass = {
                producer
                for registered_type, producer in self.producer_types.items()
                if issubclass(object_type, registered_type)
            }
            if len(producers_compatible_superclass) == 1:
                return producers_compatible_superclass.pop()
            elif len(producers_compatible_superclass) > 1:
                raise ValueError(
                    f"Type {object_type} is subclassing more than one type, this it has multiple producers within this registery: {producers_compatible_superclass}."
                )

        raise KeyError(
            f"No producer is registered for object type {object_type}. You can register a producer for specific types by subclassing `BaseProducer` and setting its `ASSOCIATED_TYPES` attribute."
        )


C = TypeVar("C", bound=BaseProducer)
producer_registry = ProducerRegistry()


def register_producer_class(cls: Type[C]) -> Type[C]:
    """Registers the producer class and returns it unmodified."""
    if not cls.TYPES:
        raise ProducerInterfaceError(
            f"Invalid producer. When defining producer, make sure to specify at least 1 type in the TYPES class variable."
        )

    for artifact_type in cls.ARTIFACT_TYPES:
        if not (
            isclass(artifact_type) and issubclass(artifact_type, BaseArtifact)
        ):
            raise ProducerInterfaceError(
                f"Associated artifact type {artifact_type} for producer is not a class or is not a subclass of BaseArtifact."
            )

    artifact_types = cls.ARTIFACT_TYPES or (BaseArtifact,)
    for t in cls.TYPES:
        if not isclass(t):
            raise ProducerInterfaceError(
                f"Associated type {t} for producer is not a class."
            )

        producer_registry.register_producer(
            t,
            cls,
        )
        type_registry.register_artifact_type(
            t,
            artifact_types,
        )

    return cls



"""
## Experiment Trackers

Experiment trackers help you track/log ML experiments by logging the
parameters, metrics, and model artifacts which allows you to compare
between runs. In general a single pipeline run is an experiment however,
one can run a series of experiments within a single pipeline.
"""

from coalescenceml.experiment_tracker.base_experiment_tracker import (
    BaseExperimentTracker,
)

__all__ = [
    "BaseExperimentTracker",
]



from __future__ import annotations

import inspect
from abc import abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    NoReturn,
    Optional,
    Set,
    Text,
    Tuple,
    Type,
    TypeVar,
    cast,
)

from coalescenceml.config.config_keys import (
    PipelineConfigurationKeys,
    StepConfigurationKeys,
)
from coalescenceml.constants import (
    ENV_COML_PREVENT_PIPELINE_EXECUTION,
    SHOULD_PREVENT_PIPELINE_EXECUTION,
)
from coalescenceml.directory import Directory
from coalescenceml.integrations.registry import integration_registry
from coalescenceml.io import fileio, utils
from coalescenceml.logger import get_logger
from coalescenceml.pipeline.exceptions import (
    PipelineConfigurationError,
    PipelineInterfaceError,
)
from coalescenceml.pipeline.runtime_configuration import RuntimeConfiguration
from coalescenceml.pipeline.schedule import Schedule
from coalescenceml.stack.exceptions import StackValidationError
from coalescenceml.step import BaseStep
from coalescenceml.utils import yaml_utils


if TYPE_CHECKING:
    from coalescenceml.stack import Stack

logger = get_logger(__name__)
PIPELINE_INNER_FUNC_NAME: str = "build_dag"
PARAM_ENABLE_CACHE: str = "enable_cache"
PARAM_REQUIRED_INTEGRATIONS: str = "required_integrations"
PARAM_REQUIREMENTS_FILE: str = "requirements_file"
PARAM_DOCKERIGNORE_FILE: str = "dockerignore_file"
INSTANCE_CONFIGURATION = "INSTANCE_CONFIGURATION"
PARAM_SECRETS: str = "secrets"


class BasePipelineMeta(type):
    """Pipeline Metaclass responsible for validating the pipeline definition."""

    def __new__(
        mcs, name: str, bases: Tuple[Type[Any], ...], dct: Dict[str, Any]
    ) -> "BasePipelineMeta":
        """Saves argument names for later verification purposes"""
        cls = cast(Type["BasePipeline"], super().__new__(mcs, name, bases, dct))

        cls.STEP_SPEC = {}

        build_dag_spec = inspect.getfullargspec(
            inspect.unwrap(getattr(cls, PIPELINE_INNER_FUNC_NAME))
        )
        build_dag_args = build_dag_spec.args

        if build_dag_args and build_dag_args[0] == "self":
            build_dag_args.pop(0)

        for arg in build_dag_args:
            arg_type = build_dag_spec.annotations.get(arg, None)
            cls.STEP_SPEC.update({arg: arg_type})
        return cls


T = TypeVar("T", bound="BasePipeline")


class BasePipeline(metaclass=BasePipelineMeta):
    """Abstract base class for all CoML pipelines.
    Attributes:
        name: The name of this pipeline.
        enable_cache: A boolean indicating if caching is enabled for this
            pipeline.
        requirements_file: Optional path to a pip requirements file that
            contains all requirements to run the pipeline.
        required_integrations: Optional set of integrations that need to be
            installed for this pipeline to run.
    """

    STEP_SPEC: ClassVar[Dict[str, Any]] = None  # type: ignore[assignment]

    INSTANCE_CONFIGURATION: Dict[Text, Any] = {}

    def __init__(self, *args: BaseStep, **kwargs: Any) -> None:
        kwargs.update(getattr(self, INSTANCE_CONFIGURATION))
        self.enable_cache = kwargs.pop(PARAM_ENABLE_CACHE, True)
        self.required_integrations = kwargs.pop(PARAM_REQUIRED_INTEGRATIONS, ())
        self.requirements_file = kwargs.pop(PARAM_REQUIREMENTS_FILE, None)
        self.dockerignore_file = kwargs.pop(PARAM_DOCKERIGNORE_FILE, None)
        self.secrets = kwargs.pop(PARAM_SECRETS, [])

        self.name = self.__class__.__name__
        logger.info("Creating run for pipeline: `%s`", self.name)
        logger.info(
            f'Cache {"enabled" if self.enable_cache else "disabled"} for '
            f"pipeline `{self.name}`"
        )

        self.__steps: Dict[str, BaseStep] = {}
        self._verify_arguments(*args, **kwargs)

    def _verify_arguments(self, *steps: BaseStep, **kw_steps: BaseStep) -> None:
        """Verifies the initialization args and kwargs of this pipeline.
        This method makes sure that no missing/unexpected arguments or
        arguments of a wrong type are passed when creating a pipeline. If
        all arguments are correct, saves the steps to `self.__steps`.
        Args:
            *steps: The args passed to the init method of this pipeline.
            **kw_steps: The kwargs passed to the init method of this pipeline.
        Raises:
            PipelineInterfaceError: If there are too many/few arguments or
                arguments with a wrong name/type.
        """
        input_step_keys = list(self.STEP_SPEC.keys())
        if len(steps) > len(input_step_keys):
            raise PipelineInterfaceError(
                f"Too many input steps for pipeline '{self.name}'. "
                f"This pipeline expects {len(input_step_keys)} step(s) "
                f"but got {len(steps) + len(kw_steps)}."
            )

        combined_steps = {}
        step_classes: Dict[Type[BaseStep], str] = {}

        for i, step in enumerate(steps):
            step_class = type(step)
            key = input_step_keys[i]

            if not isinstance(step, BaseStep):
                raise PipelineInterfaceError(
                    f"Wrong argument type (`{step_class}`) for positional "
                    f"argument {i} of pipeline '{self.name}'. Only "
                    f"`@step` decorated functions or instances of `BaseStep` "
                    f"subclasses can be used as arguments when creating "
                    f"a pipeline."
                )

            if step_class in step_classes:
                previous_key = step_classes[step_class]
                raise PipelineInterfaceError(
                    f"Found multiple step objects of the same class "
                    f"(`{step_class}`) for arguments '{previous_key}' and "
                    f"'{key}' in pipeline '{self.name}'. Only one step object "
                    f"per class is allowed inside a CoML pipeline."
                )

            step.pipeline_parameter_name = key
            combined_steps[key] = step
            step_classes[step_class] = key

        for key, step in kw_steps.items():
            step_class = type(step)

            if key in combined_steps:
                # a step for this key was already set by
                # the positional input steps
                raise PipelineInterfaceError(
                    f"Unexpected keyword argument '{key}' for pipeline "
                    f"'{self.name}'. A step for this key was "
                    f"already passed as a positional argument."
                )

            if not isinstance(step, BaseStep):
                raise PipelineInterfaceError(
                    f"Wrong argument type (`{step_class}`) for argument "
                    f"'{key}' of pipeline '{self.name}'. Only "
                    f"`@step` decorated functions or instances of `BaseStep` "
                    f"subclasses can be used as arguments when creating "
                    f"a pipeline."
                )

            if step_class in step_classes:
                previous_key = step_classes[step_class]
                raise PipelineInterfaceError(
                    f"Found multiple step objects of the same class "
                    f"(`{step_class}`) for arguments '{previous_key}' and "
                    f"'{key}' in pipeline '{self.name}'. Only one step object "
                    f"per class is allowed inside a CoML pipeline."
                )

            step.pipeline_parameter_name = key
            combined_steps[key] = step
            step_classes[step_class] = key

        # check if there are any missing or unexpected steps
        expected_steps = set(self.STEP_SPEC.keys())
        actual_steps = set(combined_steps.keys())
        missing_steps = expected_steps - actual_steps
        unexpected_steps = actual_steps - expected_steps

        if missing_steps:
            raise PipelineInterfaceError(
                f"Missing input step(s) for pipeline "
                f"'{self.name}': {missing_steps}."
            )

        if unexpected_steps:
            raise PipelineInterfaceError(
                f"Unexpected input step(s) for pipeline "
                f"'{self.name}': {unexpected_steps}. This pipeline "
                f"only requires the following steps: {expected_steps}."
            )

        self.__steps = combined_steps

    @abstractmethod
    def build_dag(self, *args: BaseStep, **kwargs: BaseStep) -> None:
        """Function that builds dag from the pipeline steps."""
        raise NotImplementedError

    @property
    def requirements(self) -> Set[str]:
        """Set of python requirements of this pipeline.
        This property is a combination of the requirements of
        - required integrations for this pipeline
        - the `requirements_file` specified for this pipeline
        """
        requirements = set()

        for integration_name in self.required_integrations:
            try:
                integration_requirements = (
                    integration_registry.select_integration_requirements(
                        integration_name
                    )
                )
                requirements.update(integration_requirements)
            except KeyError as e:
                raise KeyError(
                    f"Unable to find requirements for integration "
                    f"'{integration_name}'."
                ) from e

        if self.requirements_file and fileio.exists(self.requirements_file):
            with fileio.open(self.requirements_file, "r") as f:
                requirements.update(
                    {
                        requirement.strip()
                        for requirement in f.read().split("\n")
                    }
                )

        return requirements

    @property
    def steps(self) -> Dict[str, BaseStep]:
        """Returns a dictionary of pipeline steps."""
        return self.__steps

    @steps.setter
    def steps(self, steps: Dict[str, BaseStep]) -> NoReturn:
        """Setting the steps property is not allowed. This method always
        raises a PipelineInterfaceError.
        """
        raise PipelineInterfaceError("Cannot set steps manually!")

    def validate_stack(self, stack: "Stack") -> None:
        """Validates if a stack is able to run this pipeline."""
        available_step_operators = (
            {stack.step_operator.name} if stack.step_operator else set()
        )

        for step in self.steps.values():
            if (
                step.custom_step_operator
                and step.custom_step_operator not in available_step_operators
            ):
                raise StackValidationError(
                    f"Step '{step.name}' requires custom step operator "
                    f"'{step.custom_step_operator}' which is not configured in "
                    f"the active stack. Available step operators: "
                    f"{available_step_operators}."
                )

    def _reset_step_flags(self) -> None:
        """Reset the _has_been_called flag at the beginning of a pipeline run,
        to make sure a pipeline instance can be called more than once."""
        for step in self.steps.values():
            step._has_been_called = False

    def run(
        self,
        *,
        run_name: Optional[str] = None,
        schedule: Optional[Schedule] = None,
        **additional_parameters: Any,
    ) -> Any:
        """Runs the pipeline on the active stack of the current directory.
        Args:
            run_name: Name of the pipeline run.
            schedule: Optional schedule of the pipeline.
        """
        if SHOULD_PREVENT_PIPELINE_EXECUTION:
            # An environment variable was set to stop the execution of
            # pipelines. This is done to prevent execution of module-level
            # pipeline.run() calls inside docker containers which should only
            # run a single step.
            logger.info(
                "Preventing execution of pipeline '%s'. If this is not "
                "intended behavior, make sure to unset the environment "
                "variable '%s'.",
                self.name,
                ENV_COML_PREVENT_PIPELINE_EXECUTION,
            )
            return

        # Activating the built-in integrations through lazy loading
        from coalescenceml.integrations.registry import integration_registry

        integration_registry.activate_integrations()

        # Path of the file where pipeline.run() was called. This is needed by
        # the airflow orchestrator so it knows which file to copy into the DAG
        # directory
        dag_filepath = utils.resolve_relative_path(
            inspect.currentframe().f_back.f_code.co_filename
        )
        runtime_configuration = RuntimeConfiguration(
            run_name=run_name,
            dag_filepath=dag_filepath,
            schedule=schedule,
            **additional_parameters,
        )
        stack = Directory().active_stack

        stack_metadata = {
            component_flavor.value: component.FLAVOR
            for component_flavor, component in stack.components.items()
        }

        self._reset_step_flags()
        self.validate_stack(stack)

        return stack.deploy_pipeline(
            self, runtime_configuration=runtime_configuration
        )

    def with_config(
        self: T, config_file: str, overwrite_step_parameters: bool = False
    ) -> T:
        """Configures this pipeline using a yaml file.
        Args:
            config_file: Path to a yaml file which contains configuration
                options for running this pipeline.
            overwrite_step_parameters: If set to `True`, values from the
                configuration file will overwrite configuration parameters
                passed in code.
        Returns:
            The pipeline object that this method was called on.
        """
        config_yaml = yaml_utils.read_yaml(config_file)

        if PipelineConfigurationKeys.STEPS in config_yaml:
            self._read_config_steps(
                config_yaml[PipelineConfigurationKeys.STEPS],
                overwrite=overwrite_step_parameters,
            )

        return self

    def _read_config_steps(
        self, steps: Dict[str, Dict[str, Any]], overwrite: bool = False
    ) -> None:
        """Reads and sets step parameters from a config file.
        Args:
            steps: Maps step names to dicts of parameter names and values.
            overwrite: If `True`, overwrite previously set step parameters.
        """
        for step_name, step_dict in steps.items():
            StepConfigurationKeys.key_check(step_dict)

            if step_name not in self.__steps:
                raise PipelineConfigurationError(
                    f"Found '{step_name}' step in configuration yaml but it "
                    f"doesn't exist in the pipeline steps "
                    f"{list(self.__steps.keys())}."
                )

            step = self.__steps[step_name]
            step_parameters = (
                step.CONFIG_CLASS.__fields__.keys() if step.CONFIG_CLASS else {}
            )
            parameters = step_dict.get(StepConfigurationKeys.PARAMETERS_, {})
            for parameter, value in parameters.items():
                if parameter not in step_parameters:
                    raise PipelineConfigurationError(
                        f"Found parameter '{parameter}' for '{step_name}' step "
                        f"in configuration yaml but it doesn't exist in the "
                        f"configuration class `{step.CONFIG_CLASS}`. Available "
                        f"parameters for this step: "
                        f"{list(step_parameters)}."
                    )

                previous_value = step.PARAM_SPEC.get(parameter, None)

                if overwrite:
                    step.PARAM_SPEC[parameter] = value
                else:
                    step.PARAM_SPEC.setdefault(parameter, value)

                if overwrite or not previous_value:
                    logger.debug(
                        "Setting parameter %s=%s for step '%s'.",
                        parameter,
                        value,
                        step_name,
                    )
                if previous_value and not overwrite:
                    logger.warning(
                        "Parameter '%s' from configuration yaml will NOT be "
                        "set as a configuration object was given when "
                        "creating the step. Set `overwrite_step_parameters="
                        "True` when setting the configuration yaml to always "
                        "use the options specified in the yaml file.",
                        parameter,
                    )



import os
import tempfile
from typing import Any, Type

import xgboost as xgb

from coalescenceml.artifacts import DataArtifact
from coalescenceml.io import fileio
from coalescenceml.producers import BaseProducer
from coalescenceml.producers.producer_registry import register_producer_class


DEFAULT_FILENAME = "data.xgb.binary"


@register_producer_class
class XgboostDMatrixProducer(BaseProducer):
    """Producer to read and write from xgboost.DMatrix."""

    TYPES = (xgb.DMatrix,)
    ARTIFACT_TYPES = (DataArtifact,)

    def handle_input(self, data_type: Type[Any]) -> xgb.DMatrix:
        """Read XGB DMatrix data from binary file.

        Args:
            data_type: Data type to be processed.

        Returns:
            XGB DMatrix data
        """
        super().handle_input(data_type)
        filepath = os.path.join(self.artifact.uri, DEFAULT_FILENAME)

        with tempfile.TemporaryDirectory() as temp_dir:
            # Create temp file
            temp_file = os.path.join(temp_dir, DEFAULT_FILENAME)
            # Copy from artifact store to temp file
            fileio.copy(filepath, temp_file)
            matrix = xgb.DMatrix(temp_file)

        return matrix

    def handle_return(self, matrix: xgb.DMatrix) -> None:
        """Create binary data file for DMatrix.

        Args:
            matrix: An xgboost DMatrix dataset
        """
        super().handle_return(matrix)

        filepath = os.path.join(self.artifact.uri, DEFAULT_FILENAME)

        # Make temp artifact
        with tempfile.NamedTemporaryFile(mode="wb", delete=True) as temp_file:
            matrix.save_binary(temp_file.name)

            # copy file
            fileio.copy(temp_file.name, filepath)



"""
The stack store defines exactly where and how stacks are persisted across their 
life.
"""

from coalescenceml.stack_store.base_stack_store import BaseStackStore
from coalescenceml.stack_store.local_stack_store import LocalStackStore
from coalescenceml.stack_store.sql_stack_store import SqlStackStore


__all__ = [
    "BaseStackStore",
    "LocalStackStore",
    "SqlStackStore",
]



from coalescenceml.integrations.constants import WANDB
from coalescenceml.integrations.integration import Integration


class WandbIntegration(Integration):
    """Integration for Wandb."""

    NAME = WANDB
    REQUIREMENTS = ["wandb==0.12.15"]

    @classmethod
    def activate(cls) -> None:
        from coalescenceml.integrations.wandb import experiment_tracker


WandbIntegration.check_installation()



from coalescenceml.integrations.xgboost.producers.xgboost_booster_producer import (
    XgboostBoosterProducer,
)
from coalescenceml.integrations.xgboost.producers.xgboost_dmatrix_producer import (
    XgboostDMatrixProducer,
)



from coalescenceml.integrations.tensorflow.step.tensorflow_trainstep import (
    TFClassifierConfig,
    TFClassifierTrainStep,
)

__all__ = [
    "TFClassifierConfig",
    "TFClassifierTrainStep",
]



from typing import List

from pydantic import BaseModel

from coalescenceml.stack import Stack
from coalescenceml.stack_store.model import StackComponentWrapper


class StackWrapper(BaseModel):
    """Network Serializable Wrapper describing a Stack."""

    name: str
    components: List[StackComponentWrapper]

    @classmethod
    def from_stack(cls, stack: Stack) -> "StackWrapper":
        return cls(
            name=stack.name,
            components=[
                StackComponentWrapper.from_component(component)
                for t, component in stack.components.items()
            ],
        )
